import abc
import collections.abc

import typing_extensions as typing

from ..fdl import Deserializer, Serializer

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List, Native

T = typing.TypeVar("T", bound="Native", contravariant=True)


class Constraint(typing.Generic[T], metaclass=abc.ABCMeta):
    """
    This class serves as a generic super class for all constraints
    that enforce specific rules on the values they validate. Any
    subclass must implement the `validate` method to define the
    validation logic for the specific constraint.
    """

    @abc.abstractmethod
    async def validate(self, value: T) -> bool:
        """
        Validate the provided value against the constraint's rules.

        Args:
          value: The value to be validated.

        Returns:
          True if the value is valid according to the constraint.

        Raises:
          TypeError: If the provided value is not in the correct type.
          ValueError: If the provided value violates the constraints.
        """

    @abc.abstractmethod
    def serialize(self, serializer: Serializer) -> None:
        """
        Serialize the SiLA constraint into the xml-based feature language
        definition.

        Args:
          serializer: The serializer instance used to write xml tokens.
        """

    @classmethod
    @abc.abstractmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        """
        Deserialize the xml-based feature language definition into SiLA
        constraints.

        Args:
          deserializer: The deserializer instance used to read xml
            tokens.

        Raises:
          ValueError: If an invalid or unexpected token is detected.
        """



import collections.abc
import dataclasses

import typing_extensions as typing

from sila import datetime

from ..data_types.date import Date
from ..data_types.integer import Integer
from ..data_types.real import Real
from ..data_types.time import Time
from ..data_types.timestamp import Timestamp
from ..fdl import Deserializer, Serializer
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", int, float, datetime.date, datetime.time, datetime.datetime)


@dataclasses.dataclass
class MinimalExclusive(Constraint[T]):
    """
    A constraint that enforces a lower exclusive bound on a value.

    This constraint ensures that a given value is strictly greater
    than the specified minimum exclusive bound. It can be applied to
    various types, including Integer, Real, Date, Time, and
    Timestamp. The constraint raises errors if the input value is
    less than or equal to the lower bound or if the input is of an
    incorrect type.

    Args:
      value: The lower exclusive limit for the value. The value must
        be of the same type as this attribute and must be strictly
        greater than this limit.
    """

    value: T
    """
    The lower exclusive limit for the value. The value must be of
    the same type as this attribute and must be strictly greater
    than this limit.
    """

    async def validate(self, value: T) -> bool:
        if not isinstance(value, type(self.value)):
            raise TypeError(f"Expected value of type '{type(self.value).__name__}', received '{type(value).__name__}'.")

        if not value > self.value:
            raise ValueError(
                f"Value '{value}' must be strictly greater than the minimal exclusive limit of '{self.value}'."
            )

        return True

    def serialize(self, serializer: Serializer) -> None:
        if isinstance(self.value, int):
            serializer.write_int("MinimalExclusive", self.value)
        elif isinstance(self.value, float):
            serializer.write_float("MinimalExclusive", self.value)
        elif isinstance(self.value, datetime.date):
            serializer.write_date("MinimalExclusive", self.value)
        elif isinstance(self.value, datetime.time):
            serializer.write_time("MinimalExclusive", self.value)
        elif isinstance(self.value, datetime.datetime):
            serializer.write_datetime("MinimalExclusive", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (Integer, Real, Date, Time, Timestamp)):
            raise ValueError(
                f"Expected constraint's data type to be 'Integer', 'Real', 'Date', 'Time' or 'Timestamp', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="MinimalExclusive")
        if issubclass(data_type, Integer):
            value = yield from deserializer.read_int()
        elif issubclass(data_type, Real):
            value = yield from deserializer.read_float()
        elif issubclass(data_type, Date):
            value = yield from deserializer.read_date()
        elif issubclass(data_type, Time):
            value = yield from deserializer.read_time()
        elif issubclass(data_type, Timestamp):
            value = yield from deserializer.read_datetime()
        yield from deserializer.read_end_element(name="MinimalExclusive")

        return cls(typing.cast(T, value))



from .cancel_request import CancelRequest
from .client_message import ClientMessage
from .cloud_metadata import CloudMetadata
from .command_confirmation_response import CommandConfirmationResponse
from .command_execution_request import CommandExecutionRequest
from .command_execution_response import CommandExecutionResponse
from .command_parameter import CommandParameter
from .command_response_request import CommandResponseRequest
from .create_binary_upload_request import CreateBinaryUploadRequest
from .metadata_request import MetadataRequest
from .metadata_response import MetadataResponse
from .observable_command_response import ObservableCommandResponse
from .property_request import PropertyRequest
from .property_response import PropertyResponse
from .server_message import ServerMessage
from .unobservable_command_response import UnobservableCommandResponse

__all__ = [
    "CancelRequest",
    "ClientMessage",
    "CloudMetadata",
    "CommandConfirmationResponse",
    "CommandExecutionRequest",
    "CommandExecutionResponse",
    "CommandParameter",
    "CommandResponseRequest",
    "CreateBinaryUploadRequest",
    "MetadataRequest",
    "MetadataResponse",
    "ObservableCommandResponse",
    "PropertyRequest",
    "PropertyResponse",
    "ServerMessage",
    "UnobservableCommandResponse",
]



import typing_extensions as typing

ChannelOptions = typing.TypedDict(
    "ChannelOptions",
    {
        "grpc.census": typing.Annotated[
            bool,
            typing.Doc("""Whether to enable census for tracing and stats collection."""),
        ],
        "grpc.loadreporting": typing.Annotated[
            bool,
            typing.Doc("""Whether to enable load reporting."""),
        ],
        "grpc.server_call_metric_recording": typing.Annotated[
            bool,
            typing.Doc("""Whether to enable call metric recording."""),
        ],
        "grpc.minimal_stack": typing.Annotated[
            bool,
            typing.Doc("""
                Request that optional features default to off (regardless of what
                they usually default to) - to enable tight control over what gets
                enabled
            """),
        ],
        "grpc.max_concurrent_streams": typing.Annotated[
            int,
            typing.Doc("""Maximum number of concurrent incoming streams to allow on a http2 connection."""),
        ],
        "grpc.max_receive_message_length": typing.Annotated[
            int,
            typing.Doc("""Maximum message length in bytes that the channel can receive. -1 means unlimited."""),
        ],
        "grpc.max_send_message_length": typing.Annotated[
            int,
            typing.Doc("""Maximum message length in bytes that the channel can send. -1 means unlimited."""),
        ],
        "grpc.max_connection_idle_ms": typing.Annotated[
            int,
            typing.Doc(
                """Maximum time in milliseconds that a channel may have no outstanding rpcs, after which the server will close the connection. INT_MAX means unlimited."""
            ),
        ],
        "grpc.max_connection_age_ms": typing.Annotated[
            int,
            typing.Doc("""Maximum time in milliseconds that a channel may exist. INT_MAX means unlimited."""),
        ],
        "grpc.max_connection_age_grace_ms": typing.Annotated[
            int,
            typing.Doc(
                """Grace period in milliseconds after the channel reaches its max age. INT_MAX means unlimited."""
            ),
        ],
        "grpc.client_idle_timeout_ms": typing.Annotated[
            int,
            typing.Doc(
                """Timeout in milliseconds after the last RPC finishes on the client channel at which the channel goes back into IDLE state. INT_MAX means unlimited. The default value is 30 minutes and the min value is 1 second."""
            ),
        ],
        "grpc.per_message_compression": typing.Annotated[
            bool,
            typing.Doc(
                """Enable/disable support for per-message compression. Defaults to `True`, unless `grpc.minimal_stack` is enabled, in which case it defaults to `False`."""
            ),
        ],
        "grpc.per_message_decompression": typing.Annotated[
            bool,
            typing.Doc(
                """Experimental. Enable/disable support for per-message decompression. Defaults to `True`. If disabled, decompression will not be performed and the application will see the compressed message in the byte buffer."""
            ),
        ],
        "grpc.http2.initial_sequence_number": typing.Annotated[
            int,
            typing.Doc("""Initial stream ID for http2 transports."""),
        ],
        "grpc.http2.lookahead_bytes": typing.Annotated[
            int,
            typing.Doc(
                """Amount in bytes to read ahead on individual streams. Defaults to 64kb, larger values can help throughput on high-latency connections. NOTE: at some point we'd like to auto-tune this, and this parameter will become a no-op."""
            ),
        ],
        "grpc.http2.hpack_table_size.decoder": typing.Annotated[
            int,
            typing.Doc("""How much memory in bytes to use for hpack decoding."""),
        ],
        "grpc.http2.hpack_table_size.encoder": typing.Annotated[
            int,
            typing.Doc("""How much memory in bytes to use for hpack encoding."""),
        ],
        "grpc.http2.max_frame_size": typing.Annotated[
            int,
            typing.Doc(
                """How big a frame are we willing to receive via HTTP2. Min 16384, max 16777215. Larger values give lower CPU usage for large messages, but more head of line blocking for small messages."""
            ),
        ],
        "grpc.http2.bdp_probe": typing.Annotated[
            bool,
            typing.Doc("""Should BDP probing be performed?"""),
        ],
        "grpc.http2.min_ping_interval_without_data_ms": typing.Annotated[
            int,
            typing.Doc(
                """Minimum allowed time inmilliseconds between a server receiving successive ping frames without sending any data/header frame."""
            ),
        ],
        "grpc.server_max_unrequested_time_in_server": typing.Annotated[
            int,
            typing.Doc(
                """Maximum time to allow a request to be: (1) received by the server, but (2) not requested by a RequestCall (in the completion queue based API) before the request is cancelled."""
            ),
        ],
        "grpc.http2_scheme": typing.Annotated[
            int,
            typing.Doc("""Whether to override the http2 :scheme header."""),
        ],
        "grpc.http2.max_pings_without_data": typing.Annotated[
            int,
            typing.Doc(
                """How many pings can the client send before needing to send a data/header frame? (0 indicates that an infinite number of pings can be sent without sending a data frame or header frame). If experiment "max_pings_wo_data_throttle" is enabled, instead of pings being completely blocked, they are throttled."""
            ),
        ],
        "grpc.http2.max_ping_strikes": typing.Annotated[
            int,
            typing.Doc(
                """How many misbehaving pings the server can bear before sending goaway and closing the transport? (0 indicates that the server can bear an infinite number of misbehaving pings)."""
            ),
        ],
        "grpc.http2.write_buffer_size": typing.Annotated[
            int,
            typing.Doc(
                """How much data are we willing to queue up per stream if GRPC_WRITE_BUFFER_HINT is set? This is an upper bound."""
            ),
        ],
        "grpc.http2.true_binary": typing.Annotated[
            bool,
            typing.Doc("""Should we allow receipt of true-binary data on http2 connections? Defaults to `True`."""),
        ],
        "grpc.experimental.http2.enable_preferred_frame_size": typing.Annotated[
            int,
            typing.Doc(
                """Experimental: determines whether the preferred crypto frame size http2 setting sent to the peer at startup. If set to 0 (false - default), the preferred frame size is not sent to the peer. Otherwise it sends a default preferred crypto frame size value of 4GB to the peer at the startup of each connection."""
            ),
        ],
        "grpc.keepalive_time_ms": typing.Annotated[
            int,
            typing.Doc(
                """After a duration of this time the client/server pings its peer to see if the transport is still alive. Int valued, milliseconds."""
            ),
        ],
        "grpc.keepalive_timeout_ms": typing.Annotated[
            int,
            typing.Doc(
                """After waiting for a duration of this time, if the keepalive ping sender does not receive the ping ack, it will close the transport. Int valued, milliseconds."""
            ),
        ],
        "grpc.keepalive_permit_without_calls": typing.Annotated[
            bool,
            typing.Doc(
                """Is it permissible to send keepalive pings from the client without any outstanding streams. Int valued, 0(false)/1(true)."""
            ),
        ],
        "grpc.default_authority": typing.Annotated[
            str, typing.Doc("""Default authority to pass if none specified on call construction. A string.""")
        ],
        "grpc.primary_user_agent": typing.Annotated[
            str,
            typing.Doc(
                """Primary user agent: goes at the start of the user-agent metadata sent on each request. A string."""
            ),
        ],
        "grpc.secondary_user_agent": typing.Annotated[
            str,
            typing.Doc(
                """Secondary user agent: goes at the end of the user-agent metadata sent on each request. A string."""
            ),
        ],
        "grpc.min_reconnect_backoff_ms": typing.Annotated[
            int, typing.Doc("""The minimum time between subsequent connection attempts, in ms""")
        ],
        "grpc.max_reconnect_backoff_ms": typing.Annotated[
            int, typing.Doc("""The maximum time between subsequent connection attempts, in ms""")
        ],
        "grpc.initial_reconnect_backoff_ms": typing.Annotated[
            int, typing.Doc("""The time between the first and second connection attempts, in ms""")
        ],
        "grpc.dns_min_time_between_resolutions_ms": typing.Annotated[
            int, typing.Doc("""Minimum amount of time between DNS resolutions, in ms""")
        ],
        "grpc.server_handshake_timeout_ms": typing.Annotated[
            int,
            typing.Doc(
                """The timeout used on servers for finishing handshaking on an incoming connection. Defaults to 120 seconds."""
            ),
        ],
        "grpc.ssl_target_name_override": typing.Annotated[
            str,
            typing.Doc(
                """This *should* be used for testing only. The caller of the secure_channel_create functions may override the target name used for SSL host name checking using this channel argument which is of typeGRPC_ARG_STRING. If this argument is not specified, the name used for SSL host name checking will be the target parameter (assuming that the secure channel is an SSL channel). If this parameter is specified and the underlying is not an SSL channel, it will just be ignored."""
            ),
        ],
        "grpc.ssl_session_cache": typing.Annotated[
            object,
            typing.Doc(
                """If non-zero, a pointer to a session cache (a pointer of type grpc_ssl_session_cache*). (use grpc_ssl_session_cache_arg_vtable() to fetch an appropriate pointer arg vtable)"""
            ),
        ],
        "grpc.tsi.max_frame_size": typing.Annotated[
            bool, typing.Doc("""If non-zero, it will determine the maximum frame size used by TSI's frame protector.""")
        ],
        "grpc.max_metadata_size": typing.Annotated[
            int,
            typing.Doc(
                """Maximum metadata size (soft limit), in bytes. Note this limit applies to the max sum of all metadata key-value entries in a batch of headers. Some random sample of requests between this limit and `GRPC_ARG_ABSOLUTE_MAX_METADATA_SIZE` will be rejected. Defaults to maximum of 8 KB and `GRPC_ARG_ABSOLUTE_MAX_METADATA_SIZE` * 0.8 (if set)."""
            ),
        ],
        "grpc.absolute_max_metadata_size": typing.Annotated[
            int,
            typing.Doc(
                """Maximum metadata size (hard limit), in bytes. Note this limit applies to the max sum of all metadata key-value entries in a batch of headers. All requests exceeding this limit will be rejected. Defaults to maximum of 16 KB and `GRPC_ARG_MAX_METADATA_SIZE` * 1.25 (if set)."""
            ),
        ],
        "grpc.so_reuseport": typing.Annotated[
            bool, typing.Doc("""If non-zero, allow the use of SO_REUSEPORT if it's available (default 1)""")
        ],
        "grpc.resource_quota": typing.Annotated[
            bool,
            typing.Doc(
                """If non-zero, a pointer to a buffer pool (a pointer of type grpc_resource_quota*). (use grpc_resource_quota_arg_vtable() to fetch an appropriate pointer arg vtable)"""
            ),
        ],
        "grpc.expand_wildcard_addrs": typing.Annotated[
            bool, typing.Doc("""If non-zero, expand wildcard addresses to a list of local addresses.""")
        ],
        "grpc.service_config": typing.Annotated[
            str,
            typing.Doc(
                """Service config data in JSON form. This value will be ignored if the name resolver returns a service config."""
            ),
        ],
        "grpc.service_config_disable_resolution": typing.Annotated[
            bool, typing.Doc("""Disable looking up the service config via the name resolver.""")
        ],
        "grpc.lb_policy_name": typing.Annotated[str, typing.Doc("""LB policy name.""")],
        "grpc.lb.ring_hash.ring_size_cap": typing.Annotated[
            int,
            typing.Doc(
                """Cap for ring size in the ring_hash LB policy.  The min and max ring size values set in the LB policy config will be capped to this value. Default is 4096."""
            ),
        ],
        "grpc.socket_mutator": typing.Annotated[
            object, typing.Doc("""The grpc_socket_mutator instance that set the socket options. A pointer.""")
        ],
        "grpc.socket_factory": typing.Annotated[
            object, typing.Doc("""The grpc_socket_factory instance to create and bind sockets. A pointer.""")
        ],
        "grpc.max_channel_trace_event_memory_per_node": typing.Annotated[
            int,
            typing.Doc(
                """The maximum amount of memory used by trace events per channel trace node. Once the maximum is reached, subsequent events will evict the oldest events from the buffer. The unit for this knob is bytes. Setting it to zero causes channel tracing to be disabled."""
            ),
        ],
        "grpc.enable_channelz": typing.Annotated[
            bool,
            typing.Doc(
                """If non-zero, gRPC library will track stats and information at at per channel level. Disabling channelz naturally disables channel tracing. The default is for channelz to be enabled."""
            ),
        ],
        "grpc.use_cronet_packet_coalescing": typing.Annotated[
            bool, typing.Doc("""If non-zero, Cronet transport will coalesce packets to fewer frames when possible.""")
        ],
        "grpc.experimental.tcp_read_chunk_size": typing.Annotated[
            int,
            typing.Doc(
                """Channel arg (integer) setting how large a slice to try and read from the wire each time recvmsg (or equivalent) is called. Defaults to 8191"""
            ),
        ],
        "grpc.experimental.tcp_min_read_chunk_size": typing.Annotated[int, typing.Doc("""""")],
        "grpc.experimental.tcp_max_read_chunk_size": typing.Annotated[int, typing.Doc("""""")],
        "grpc.experimental.tcp_tx_zerocopy_enabled": typing.Annotated[
            bool,
            typing.Doc(
                """TCP TX Zerocopy enable state: zero is disabled, non-zero is enabled. By default, it is disabled."""
            ),
        ],
        "grpc.experimental.tcp_tx_zerocopy_send_bytes_threshold": typing.Annotated[
            int,
            typing.Doc(
                """TCP TX Zerocopy send threshold: only zerocopy if >= this many bytes sent. By default, this is set to 16KB."""
            ),
        ],
        "grpc.experimental.tcp_tx_zerocopy_max_simultaneous_sends": typing.Annotated[
            int,
            typing.Doc(
                """TCP TX Zerocopy max simultaneous sends: limit for maximum number of pending calls to tcp_write() using zerocopy. A tcp_write() is considered pending until the kernel performs the zerocopy-done callback for all sendmsg() calls issued by the tcp_write(). By default, this is set to 4."""
            ),
        ],
        "grpc.tcp_receive_buffer_size": typing.Annotated[
            int, typing.Doc("""Overrides the TCP socket receive buffer size, SO_RCVBUF.""")
        ],
        "grpc.grpclb_call_timeout_ms": typing.Annotated[
            int,
            typing.Doc(
                """Timeout in milliseconds to use for calls to the grpclb load balancer. If 0 or unset, the balancer calls will have no deadline."""
            ),
        ],
        "grpc.TEST_ONLY_DO_NOT_USE_IN_PROD.xds_bootstrap_config": typing.Annotated[
            str,
            typing.Doc(
                """Specifies the xDS bootstrap config as a JSON string. FOR TESTING PURPOSES ONLY -- DO NOT USE IN PRODUCTION. This option allows controlling the bootstrap configuration on a per-channel basis, which is useful in tests.  However, this results in having a separate xDS client instance per channel rather than using the global instance, which is not the intended way to use xDS. Currently, this will (a) add unnecessary load on the xDS server and (b) break use of CSDS, and there may be additional side effects in the future."""
            ),
        ],
        "grpc.grpclb_fallback_timeout_ms": typing.Annotated[
            int,
            typing.Doc(
                """Timeout in milliseconds to wait for the serverlist from the grpclb load balancer before using fallback backend addresses from the resolver. If 0, enter fallback mode immediately. Default value is 10000."""
            ),
        ],
        "grpc.experimental.grpclb_channel_args": typing.Annotated[
            object,
            typing.Doc(
                """Experimental Arg. Channel args to be used for the control-plane channel created to the grpclb load balancers. This is a pointer arg whose value is a grpc_channel_args object. If unset, most channel args from the parent channel will be propagated to the grpclb channel."""
            ),
        ],
        "grpc.priority_failover_timeout_ms": typing.Annotated[
            int,
            typing.Doc(
                """Timeout in milliseconds to wait for the child of a specific priority to complete its initial connection attempt before the priority LB policy fails over to the next priority. Default value is 10 seconds."""
            ),
        ],
        "grpc.workaround.cronet_compression": typing.Annotated[
            bool, typing.Doc("""If non-zero, grpc server's cronet compression workaround will be enabled""")
        ],
        "grpc.optimization_target": typing.Annotated[
            typing.Literal["latency", "blend", "throughput"],
            typing.Doc(
                """String defining the optimization target for a channel. Can be: "latency"    - attempt to minimize latency at the cost of throughput "blend"      - try to balance latency and throughput "throughput" - attempt to maximize throughput at the expense of latency Defaults to "blend". In the current implementation "blend" is equivalent to "latency"."""
            ),
        ],
        "grpc.enable_retries": typing.Annotated[
            bool,
            typing.Doc(
                """Enables retry functionality.  Defaults to true.  When enabled, transparent retries will be performed as appropriate, and configurable retries are enabled when they are configured via the service config. For details, see:   https://github.com/grpc/proposal/blob/master/A6-client-retries.md NOTE: Hedging functionality is not yet implemented, so those fields in the service config will currently be ignored.  See also the GRPC_ARG_EXPERIMENTAL_ENABLE_HEDGING arg below."""
            ),
        ],
        "grpc.experimental.enable_hedging": typing.Annotated[
            bool,
            typing.Doc(
                """Enables hedging functionality, as described in:   https://github.com/grpc/proposal/blob/master/A6-client-retries.md Default is currently false, since this functionality is not yet fully implemented. NOTE: This channel arg is experimental and will eventually be removed. Once hedging functionality has been implemented and proves stable, this arg will be removed, and the hedging functionality will be enabled via the GRPC_ARG_ENABLE_RETRIES arg above."""
            ),
        ],
        "grpc.per_rpc_retry_buffer_size": typing.Annotated[
            int, typing.Doc("""Per-RPC retry buffer size, in bytes. Default is 256 KiB.""")
        ],
        "grpc.mobile_log_context": typing.Annotated[
            object,
            typing.Doc(
                """Channel arg that carries the bridged objective c object for custom metrics logging filter."""
            ),
        ],
        "grpc.disable_client_authority_filter": typing.Annotated[
            bool, typing.Doc("""If non-zero, client authority filter is disabled for the channel""")
        ],
        "grpc.enable_http_proxy": typing.Annotated[
            bool, typing.Doc("""If set to zero, disables use of http proxies. Enabled by default.""")
        ],
        "grpc.http_proxy": typing.Annotated[
            bool,
            typing.Doc(
                """Channel arg to set http proxy per channel. If set, the channel arg value will be preferred over the environment variable settings."""
            ),
        ],
        "grpc.address_http_proxy": typing.Annotated[
            str,
            typing.Doc(
                """Specifies an HTTP proxy to use for individual addresses. The proxy must be specified as an IP address, not a DNS name. If set, the channel arg value will be preferred over the environment variable settings."""
            ),
        ],
        "grpc.address_http_proxy_enabled_addresses": typing.Annotated[
            str,
            typing.Doc(
                """Comma separated list of addresses or address ranges that are behind the address HTTP proxy."""
            ),
        ],
        "grpc.surface_user_agent": typing.Annotated[
            bool,
            typing.Doc(
                """If set to non zero, surfaces the user agent string to the server. User agent is surfaced by default."""
            ),
        ],
        "grpc.inhibit_health_checking": typing.Annotated[
            bool, typing.Doc("""If set, inhibits health checking (which may be enabled via the service config.)""")
        ],
        "grpc.dns_enable_srv_queries": typing.Annotated[
            bool,
            typing.Doc(
                """If enabled, the channel's DNS resolver queries for SRV records. This is useful only when using the "grpclb" load balancing policy, as described in the following documents:  https://github.com/grpc/proposal/blob/master/A5-grpclb-in-dns.md  https://github.com/grpc/proposal/blob/master/A24-lb-policy-config.md  https://github.com/grpc/proposal/blob/master/A26-grpclb-selection.md Note that this works only with the "ares" DNS resolver; it isn't supported by the "native" DNS resolver."""
            ),
        ],
        "grpc.dns_ares_query_timeout": typing.Annotated[
            bool,
            typing.Doc(
                """If set, determines an upper bound on the number of milliseconds that the c-ares based DNS resolver will wait on queries before cancelling them. The default value is 120,000. Setting this to "0" will disable the overall timeout entirely. Note that this doesn't include internal c-ares timeouts/backoff/retry logic, and so the actual DNS resolution may time out sooner than the value specified here."""
            ),
        ],
        "grpc.use_local_subchannel_pool": typing.Annotated[
            bool,
            typing.Doc(
                """If set, uses a local subchannel pool within the channel. Otherwise, uses the global subchannel pool."""
            ),
        ],
        "grpc.channel_pooling_domain": typing.Annotated[
            str, typing.Doc("""gRPC Objective-C channel pooling domain string.""")
        ],
        "grpc.channel_id": typing.Annotated[int, typing.Doc("""gRPC Objective-C channel pooling id.""")],
        "grpc.authorization_policy_provider": typing.Annotated[
            bool,
            typing.Doc(
                """Channel argument for grpc_authorization_policy_provider. If present, enables gRPC authorization check."""
            ),
        ],
        "grpc.experimental.server_config_change_drain_grace_time_ms": typing.Annotated[
            int,
            typing.Doc(
                """EXPERIMENTAL. Updates to a server's configuration from a config fetcher (for example, listener updates from xDS) cause all older connections to be gracefully shut down (i.e., "drained") with a grace period configured by this channel arg. Int valued, milliseconds. Defaults to 10 minutes."""
            ),
        ],
        "grpc.dscp": typing.Annotated[
            int,
            typing.Doc(
                """Configure the Differentiated Services Code Point used on outgoing packets. Integer value ranging from 0 to 63."""
            ),
        ],
        "grpc.happy_eyeballs_connection_attempt_delay_ms": typing.Annotated[
            int,
            typing.Doc("""Connection Attempt Delay for use in Happy Eyeballs, in milliseconds. Defaults to 250ms."""),
        ],
        "grpc.event_engine_use_memory_allocator_factory": typing.Annotated[
            object,
            typing.Doc(
                """It accepts a MemoryAllocatorFactory as input and If specified, it forces the default event engine to use memory allocators created using the provided factory."""
            ),
        ],
        "grpc.max_allowed_incoming_connections": typing.Annotated[
            int,
            typing.Doc(
                """Configure the max number of allowed incoming connections to the server. If unspecified, it is unlimited"""
            ),
        ],
        "grpc.experimental.stats_plugins": typing.Annotated[
            object, typing.Doc("""Configure per-channel or per-server stats plugins.""")
        ],
        "grpc.security_frame_allowed": typing.Annotated[
            bool, typing.Doc("""If non-zero, allow security frames to be sent and received.""")
        ],
    },
    total=False,
)



import collections.abc
import dataclasses
import functools
import weakref

import typing_extensions as typing

from ..data_types import Any, Convertable, Native
from ..errors import InvalidMetadata
from ..fdl import Serializable
from ..identifiers import MetadataIdentifier
from ..protobuf import ConversionError, DecodeError, Message
from ..validators import check_display_name, check_identifier

if typing.TYPE_CHECKING:
    from ..common import Feature, Handler
    from ..data_types import DataType
    from ..errors import DefinedExecutionError
    from ..fdl import Deserializer, Serializer

T = typing.TypeVar("T", bound=Native)


@dataclasses.dataclass
class Metadata(Message, Serializable, Convertable[T], typing.Generic[T]):
    """Aditional information the server expects to receive from a client."""

    identifier: typing.ClassVar[str] = ""
    display_name: typing.ClassVar[str] = ""
    description: typing.ClassVar[str] = ""
    data_type: typing.ClassVar[type["DataType"]] = Any
    errors: typing.ClassVar[dict[str, type["DefinedExecutionError"]]] = {}
    affects: typing.ClassVar[collections.abc.Sequence[str]] = []
    feature: typing.ClassVar[typing.Optional["Feature"]] = None

    value: "DataType" = dataclasses.field(default_factory=Any)

    @classmethod
    async def from_buffer(
        cls, context: "Handler", metadata: typing.Optional[dict[str, bytes]] = None, /
    ) -> typing.Self:
        """
        Convert a native Python value to its corresponding SiLA data type.

        Args:
          context: The context in which the conversion is performed.
          metadata: The received metadata dictionary.

        Returns:
          An instance of the corresponding SiLA metadata.

        Raises:
          ConversionError: If the provided native value cannot be
            converted to the SiLA metadata.
        """

        assert context.feature

        metadata = metadata or {}
        rpc_header = cls.rpc_header()

        if rpc_header not in metadata or not (metadatum := metadata[rpc_header]):
            msg = f"Missing matadata '{cls.identifier}' in {context.__class__.__name__} '{context.identifier}'."
            raise InvalidMetadata(msg)

        try:
            return cls.decode(metadatum)
        except DecodeError as error:
            msg = (
                f"Unable to decode matadata '{cls.identifier}' in "
                f"{context.__class__.__name__} '{context.identifier}': {error.message}"
            )
            raise InvalidMetadata(msg) from None

    @typing.override
    async def to_native(self, context: "Handler", /) -> T:
        assert context.feature

        try:
            return await self.value.to_native(context.feature.context)
        except ConversionError as error:
            msg = (
                f"Unable to decode matadata '{self.identifier}' in "
                f"{context.__class__.__name__} '{context.identifier}': {error.message}"
            )
            raise InvalidMetadata(msg) from None

    @classmethod
    @functools.cache
    def fully_qualified_identifier(cls) -> MetadataIdentifier:
        """Uniquely identifies the metadata."""

        if cls.feature is None:
            msg = (
                f"Unable to access fully qualified identifier of unregistered Metadata "
                f"'{cls.identifier}'. Add it to a feature first."
            )
            raise RuntimeError(msg)

        return MetadataIdentifier.create(**cls.feature.fully_qualified_identifier._data, metadata=cls.identifier)

    @classmethod
    @functools.cache
    def rpc_header(cls) -> str:
        """Get the gRPC header specifier used to identify metadata."""

        return f"sila-{cls.fully_qualified_identifier().lower().replace('/', '-')}-bin"

    @typing.override
    @classmethod
    def serialize(cls, serializer: "Serializer") -> None:
        serializer.start_element("Metadata")
        serializer.write_str("Identifier", cls.identifier)
        serializer.write_str("DisplayName", cls.display_name)
        serializer.write_str("Description", cls.description)
        cls.data_type.serialize(serializer)
        if cls.errors:
            serializer.start_element("DefinedExecutionErrors")
            for Error in cls.errors.values():
                serializer.write_str("Identifier", Error.identifier)
            serializer.end_element("DefinedExecutionErrors")
        serializer.end_element("Metadata")

    @typing.override
    @classmethod
    def deserialize(cls, deserializer: "Deserializer") -> collections.abc.Generator[None, typing.Any, typing.Self]: ...

    @classmethod
    def add_to_feature(cls, feature: "Feature") -> type[typing.Self]:
        """
        Register this metadata with a feature.

        Args:
          feature: The feature to add this metadata to.

        Returns:
          The class, allowing for method chaining.
        """

        cls.feature = weakref.proxy(feature)
        feature.metadata[cls.identifier] = cls

        return cls

    @classmethod
    def create(
        cls,
        identifier: str,
        display_name: str,
        description: str = "",
        data_type: type["DataType"] = Any,
        errors: typing.Optional[collections.abc.Mapping[str, type["DefinedExecutionError"]]] = None,
        affects: typing.Optional[collections.abc.Sequence[str]] = None,
        feature: typing.Optional["Feature"] = None,
        **kwargs,
    ) -> type[typing.Self]:
        """
        Create a new SiLA `Custom` class with the provided data type.

        Args:
          identifier: Uniquely identifies the custom data type within the
            scope of the same feature.
          display_name: Human readable name of the custom data type.
          description: Describes the use and purpose of the custom data
            type.
          data_type: The SiLA data type for the custom value.
          errors: A list of defined execution errors that can happen when
            accessing this handler.
          affects: A list of handlers affected by this metadata.
          feature: The feature custom data type is assigned to.
          name: An optional name for the new `Custom` class.

        Returns:
          A new `Custom` class with the specified data type.
        """

        check_identifier(identifier)
        check_display_name(display_name)

        metadata: type[typing.Self] = dataclasses.make_dataclass(
            identifier or cls.__name__,
            [("value", data_type, dataclasses.field(default_factory=data_type))],
            bases=(cls,),
            namespace={
                "__doc__": description,
                "identifier": identifier,
                "display_name": display_name,
                "description": description,
                "data_type": data_type,
                "errors": dict(errors or {}),
                "affects": list(affects or []),
                "feature": feature,
                **kwargs,
            },
            eq=False,
        )

        if feature is not None:
            metadata.add_to_feature(feature)

        return metadata

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Metadata):
            return NotImplemented

        return (
            self.identifier == other.identifier
            and self.display_name == other.display_name
            and self.description == other.description
            and self.data_type.__name__ == other.data_type.__name__
            and self.errors == other.errors
            and self.affects == other.affects
            and self.value == other.value
        )



import dataclasses

import typing_extensions as typing

from sila import datetime

from ..protobuf import Reader, WireType, Writer
from .data_type import BasicType
from .timezone import Timezone

if typing.TYPE_CHECKING:
    from ..common import Context


@dataclasses.dataclass
class Time(BasicType[datetime.time]):
    """
    Represents an ISO 8601 time, with an additional timezone (as an offset from UTC).

    Attributes:
      hour: The hour of the time value in range [0-23]. Defaults to
        zero.
      minute: The minute of the time value in range [0-59]. Defaults
        to zero.
      second: The second of the time value in range [0-59]. Defaults
        to zero.
      millisecond: The millisecond of the time value in range
        [0-999]. Defaults to zero.
      timezone: The timezone of the time value. Defaults to utc.
    """

    hour: int = 0
    minute: int = 0
    second: int = 0
    millisecond: int = 0
    timezone: Timezone = dataclasses.field(default_factory=Timezone)

    @typing.overload
    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[datetime.time] = None, /) -> typing.Self: ...

    @typing.overload
    @classmethod
    async def from_native(
        cls,
        context: "Context",
        /,
        *,
        hour: int = 0,
        minute: int = 0,
        second: int = 0,
        millisecond: int = 0,
        hours: int = 0,
        minutes: int = 0,
    ) -> typing.Self: ...

    @typing.override
    @classmethod
    async def from_native(
        cls,
        context: "Context",
        value: typing.Optional[datetime.time] = None,
        /,
        *,
        hour: int = 0,
        minute: int = 0,
        second: int = 0,
        millisecond: int = 0,
        hours: int = 0,
        minutes: int = 0,
    ) -> typing.Self:
        if value is None:
            return await cls(
                hour=hour,
                minute=minute,
                second=second,
                millisecond=millisecond,
                timezone=await Timezone.from_native(context, hours=hours, minutes=minutes),
            ).validate()

        return await cls(
            hour=value.hour,
            minute=value.minute,
            second=value.second,
            millisecond=int(value.microsecond / 1000),
            timezone=await Timezone.from_native(
                context,
                datetime.timezone(offset)
                if (tzinfo := getattr(value, "tzinfo", None))
                and isinstance(tzinfo, datetime.tzinfo)
                and (offset := tzinfo.utcoffset(None))
                else datetime.timezone.utc,
            ),
        ).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> datetime.time:
        await self.validate()

        return datetime.time(
            hour=int(self.hour),
            minute=int(self.minute),
            second=int(self.second),
            microsecond=int(self.millisecond) * 1000,
            tzinfo=await self.timezone.to_native(context) if self.timezone else None,
        )

    @typing.override
    async def validate(self) -> typing.Self:
        if not isinstance(self.hour, int):
            msg = f"Expected hour of type 'int', received '{type(self.hour).__name__}'."
            raise TypeError(msg)

        if not isinstance(self.minute, int):
            msg = f"Expected minute of type 'int', received '{type(self.minute).__name__}'."
            raise TypeError(msg)

        if not isinstance(self.second, int):
            msg = f"Expected second of type 'int', received '{type(self.second).__name__}'."
            raise TypeError(msg)

        if not isinstance(self.millisecond, int):
            msg = f"Expected millisecond of type 'int', received '{type(self.millisecond).__name__}'."
            raise TypeError(msg)

        if not (0 <= self.hour <= 23):
            msg = f"Hour must be between 0 and 23, received '{self.hour}'."
            raise ValueError(msg)

        if not (0 <= self.minute <= 59):
            msg = f"Minute must be between 0 and 59, received '{self.minute}'."
            raise ValueError(msg)

        if not (0 <= self.second <= 59):
            msg = f"Second must be between 0 and 59, received '{self.second}'."
            raise ValueError(msg)

        if not (0 <= self.millisecond <= 999):
            msg = f"Millisecond must be between 0 and 999, received '{self.millisecond}'."
            raise ValueError(msg)

        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                message.second = reader.read_int32()
            elif field_number == 2:
                reader.expect_type(tag, WireType.VARINT)
                message.minute = reader.read_uint32()
            elif field_number == 3:
                reader.expect_type(tag, WireType.VARINT)
                message.hour = reader.read_uint32()
            elif field_number == 4:
                reader.expect_type(tag, WireType.LEN)
                message.timezone = Timezone.decode(reader, reader.read_uint32())
            elif field_number == 5:
                reader.expect_type(tag, WireType.VARINT)
                message.millisecond = reader.read_uint32()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.second:
            writer.write_uint32(8).write_uint32(self.second)
        if self.minute:
            writer.write_uint32(16).write_uint32(self.minute)
        if self.hour:
            writer.write_uint32(24).write_uint32(self.hour)
        self.timezone.encode(writer, 4)
        if self.millisecond:
            writer.write_uint32(40).write_uint32(self.millisecond)

        if number:
            writer.ldelim()

        return writer.finish()



import enum

import typing_extensions as typing

from ..protobuf import Reader, WireType, Writer
from .sila_error import SiLAError


class FrameworkErrorType(enum.IntEnum):
    """Type of errors that occur when SiLA 2 is used incorrectly."""

    COMMAND_EXECUTION_NOT_ACCEPTED = 0
    """
    Is issued if the SiLA server does not allow the command execution
    because it is occupied handling other command executions.
    """

    INVALID_COMMAND_EXECUTION_UUID = 1
    """
    Is issued if the provided command execution UUID is invalid or
    not recognized.
    """

    COMMAND_EXECUTION_NOT_FINISHED = 2
    """Is issued if the response to an unfinished command is requested."""

    INVALID_METADATA = 3
    """Is issued if metadata is missing or invalid."""

    NO_METADATA_ALLOWED = 4
    """Is issued if providing metadata is not allowed."""


class FrameworkError(SiLAError):
    """
    Error that occurs when SiLA 2 is used incorrectly.

    Args:
      error_type: The type of error that occurred.
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    Type: typing.ClassVar[type[FrameworkErrorType]] = FrameworkErrorType

    def __init__(self, message: str, error_type: FrameworkErrorType) -> None:
        super().__init__(message, error_type)

    @property
    def error_type(self) -> FrameworkErrorType:
        """The type of error that occurred."""

        return self.args[1]

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = ""
        error_type = FrameworkErrorType.COMMAND_EXECUTION_NOT_ACCEPTED
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                error_type = FrameworkErrorType(reader.read_int32())
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message = reader.read_string()
            else:
                reader.skip_type(tag & 7)

        return cls(message, error_type)

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        writer.write_uint32(34).fork()

        if self.error_type:
            writer.write_uint32(8).write_int32(self.error_type)
        if self.message:
            writer.write_uint32(18).write_string(self.message)

        writer.ldelim()

        if number:
            writer.ldelim()

        return writer.finish()


class CommandExecutionNotAccepted(FrameworkError):
    """
    Raised when the server does not allow further command executions.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message, FrameworkErrorType.COMMAND_EXECUTION_NOT_ACCEPTED)


class InvalidCommandExecutionUUID(FrameworkError):
    """
    Raised when the provided identifier is invalid or not recognized.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message, FrameworkErrorType.INVALID_COMMAND_EXECUTION_UUID)


class CommandExecutionNotFinished(FrameworkError):
    """
    Raised when the response to an unfinished command is requested.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message, FrameworkErrorType.COMMAND_EXECUTION_NOT_FINISHED)


class InvalidMetadata(FrameworkError):
    """
    Raised when metadata is missing or invalid.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message, FrameworkErrorType.INVALID_METADATA)


class NoMetadataAllowed(FrameworkError):
    """
    Raised when providing metadata is not allowed.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message, FrameworkErrorType.NO_METADATA_ALLOWED)



import builtins
import collections.abc
import dataclasses

import typing_extensions as typing

from ..data_types.binary import Binary
from ..data_types.string import String
from ..fdl import Characters, Deserializer, EndElement, Serializer, StartElement
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", str, bytes)


class ContentTypeParameter(typing.NamedTuple):
    """
    Represents a parameter for the content type, consisting of an
    attribute and its corresponding value.

    Args:
      attribute: The name of the parameter attribute, e.g. 'charset'.
      value: The value of the parameter attribute, e.g. 'utf-8'.
    """

    attribute: str
    """The name of the parameter attribute, e.g. 'charset'."""

    value: str
    """The value of the parameter attribute, e.g. 'utf-8'."""


@dataclasses.dataclass(init=False)
class ContentType(Constraint[T]):
    """
    A constraint that defines a media type (content type) with
    optional parameters.

    This class defines the structure of a content type with its type,
    subtype, and optional parameters. It provides a property to
    generate the full media type string and validates that the input
    value is of the appropriate type (`String` or `Binary`).

    Args:
      type: The main type of the content (e.g., 'application',
        'text').
      subtype: The subtype of the content (e.g., 'json', 'html').
      parameters: A list of additional parameters for the content
         type (e.g., 'charset' or 'boundary').
    """

    Parameter: typing.ClassVar[builtins.type[ContentTypeParameter]] = ContentTypeParameter
    """
    Represents a parameter for the content type, consisting of an
    attribute and its corresponding value.
    """

    type: str
    """
    The main type of the content (e.g., 'application', 'text').
    """

    subtype: str
    """
    The subtype of the content (e.g., 'json', 'html').
    """

    parameters: list[ContentTypeParameter] = dataclasses.field(default_factory=list)
    """
    A list of additional parameters for the content type (e.g.,
    'charset' or 'boundary').
    """

    @property
    def media_type(self) -> str:
        """
        The full media type string, including the type, subtype, and any
        additional parameters in the format 'type/subtype; param=value'.
        """

        return f"{self.type}/{self.subtype}" + "".join(
            [f"; {parameter[0]}={parameter[1]}" for parameter in self.parameters]
        )

    def __init__(
        self,
        type: str,
        subtype: str,
        parameters: typing.Optional[
            collections.abc.Sequence[typing.Union[ContentTypeParameter, tuple[str, str]]]
        ] = None,
    ) -> None:
        self.type = type
        self.subtype = subtype
        self.parameters: list[ContentTypeParameter] = [
            parameter
            if isinstance(parameter, ContentTypeParameter)
            else ContentTypeParameter(parameter[0], parameter[1])
            for parameter in parameters or []
        ]

    async def validate(self, value: T) -> bool:
        if not isinstance(value, (str, bytes)):
            raise TypeError(f"Expected value of type 'str' or 'bytes', received '{type(value).__name__}'.")

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.start_element("ContentType")
        serializer.write_str("Type", self.type)
        serializer.write_str("Subtype", self.subtype)

        if len(self.parameters):
            serializer.start_element("Parameters")
            for parameter in self.parameters:
                serializer.start_element("Parameter")
                serializer.write_str("Attribute", parameter.attribute)
                serializer.write_str("Value", parameter.value)
                serializer.end_element("Parameter")
            serializer.end_element("Parameters")

        serializer.end_element("ContentType")

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[builtins.type["BasicType"], builtins.type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (String, Binary)):
            raise ValueError(
                f"Expected constraint's data type to be 'String' or 'Binary', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="ContentType")

        # Type
        yield from deserializer.read_start_element(name="Type")
        type = yield from deserializer.read_str()
        yield from deserializer.read_end_element(name="Type")

        # Subtype
        yield from deserializer.read_start_element(name="Subtype")
        subtype = yield from deserializer.read_str()
        yield from deserializer.read_end_element(name="Subtype")

        token = yield

        if isinstance(token, StartElement):
            if token.name != "Parameters":
                raise ValueError(
                    f"Expected start element with name 'Parameters', received start element with name '{token.name}'."
                )

        elif isinstance(token, Characters):
            raise ValueError(
                f"Expected start element with name 'Parameters' or end element with name 'ContentType', received characters '{token.value}'."
            )

        elif isinstance(token, EndElement) and token.name == "ContentType":
            return cls(type, subtype)

        parameters: list[ContentTypeParameter] = []
        while True:
            token = yield

            if isinstance(token, StartElement):
                if token.name == "Parameter":
                    # Attribute
                    yield from deserializer.read_start_element(name="Attribute")
                    attribute = yield from deserializer.read_str()
                    yield from deserializer.read_end_element(name="Attribute")

                    # Value
                    yield from deserializer.read_start_element(name="Value")
                    value = yield from deserializer.read_str()
                    yield from deserializer.read_end_element(name="Value")

                    parameters.append(ContentTypeParameter(attribute, value))

                else:
                    raise ValueError(
                        f"Expected start element with name 'Parameter', received start element with name '{token.name}'."
                    )

            elif isinstance(token, EndElement):
                if token.name == "Parameter":
                    continue
                else:
                    break  # pragma: no cover

            elif isinstance(token, Characters):
                raise ValueError(f"Expected start element with name 'Parameter', received characters '{token.value}'.")

        if not parameters:
            raise ValueError("Expected at least one 'Parameter' element inside the 'ContentType' element.")

        yield from deserializer.read_end_element(name="ContentType")

        return cls(type, subtype, parameters)



import collections.abc
import dataclasses
import functools

import typing_extensions as typing

from ..fdl import Serializable
from ..identifiers import FeatureIdentifier
from ..validators import check_display_name, check_identifier
from .context_proxy import ContextProxy

if typing.TYPE_CHECKING:
    from ..command import Command
    from ..data_types import Custom
    from ..errors import DefinedExecutionError
    from ..fdl import Deserializer, Serializer
    from ..metadata import Metadata
    from ..property import Property
    from .context import Context


@dataclasses.dataclass
class Feature(Serializable):
    """Describes a specific behavior of a SiLA server."""

    locale: str = dataclasses.field(default="en-us")
    sila2_version: str = dataclasses.field(default="1.1")
    version: str = dataclasses.field(default="1.0")
    maturity_level: typing.Literal["Draft", "Verified", "Normative"] = dataclasses.field(default="Draft")
    originator: str = dataclasses.field(default="org.silastandard")
    category: str = dataclasses.field(default="none")

    identifier: str = ""
    """
    A feature identifier is the identifier of a feature. Each feature must have a feature identifier. All features
    sharing the scope of the same originator and category must have unique feature identifiers. Uniqueness must be
    checked without taking lower and upper case into account.
    """

    display_name: str = ""
    """Human readable name of the SiLA feature."""

    description: str = dataclasses.field(default="")
    """
    A feature description is the description of a feature. A feature description must describe the behaviors /
    capabilities the feature models in human readable form and with as many details as possible.
    """

    commands: dict[str, "Command"] = dataclasses.field(default_factory=dict)

    properties: dict[str, "Property"] = dataclasses.field(default_factory=dict)

    metadata: dict[str, type["Metadata"]] = dataclasses.field(default_factory=dict)

    errors: dict[str, type["DefinedExecutionError"]] = dataclasses.field(default_factory=dict)

    data_type_definitions: dict[str, type["Custom"]] = dataclasses.field(default_factory=dict)

    context: "Context" = dataclasses.field(default_factory=ContextProxy)
    """The context (either client or server) this feature was registered with."""

    def __post_init__(self) -> None:
        check_identifier(self.identifier)
        check_display_name(self.display_name)

        for command in self.commands.values():
            command.feature = self

        for property_ in self.properties.values():
            property_.feature = self

        for metadata in self.metadata.values():
            metadata.add_to_feature(self)

    @functools.cached_property
    def fully_qualified_identifier(self) -> FeatureIdentifier:
        """Uniquely identifies the feature."""

        return FeatureIdentifier.create(
            self.originator, self.category, self.identifier, int(self.version.rpartition(".")[0])
        )

    @functools.cached_property
    def rpc_package(self) -> str:
        """The package specifier to namespace services and protobuf messages."""

        return ".".join(
            (
                "sila2",
                self.originator,
                self.category,
                str(self.identifier).lower(),
                f"v{self.version.rpartition('.')[0]}",
            )
        )

    @typing.override
    def serialize(self, serializer: "Serializer") -> None:
        serializer.write(
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            f'<Feature Locale="{self.locale}" SiLA2Version="{self.sila2_version}" FeatureVersion="{self.version}" '
            f'MaturityLevel="{self.maturity_level}" Originator="{self.originator}" Category="{self.category}"\n'
            '         xmlns="http://www.sila-standard.org"\n'
            '         xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"\n'
            '         xsi:schemaLocation="http://www.sila-standard.org https://gitlab.com/SiLA2/sila_base/raw/master/schema/FeatureDefinition.xsd">'
        )
        serializer.indent()
        serializer.write_str("Identifier", self.identifier)
        serializer.write_str("DisplayName", self.display_name)
        serializer.write_str("Description", self.description)

        defined_execution_errors: dict[str, type["DefinedExecutionError"]] = self.errors.copy()

        for command in self.commands.values():
            command.serialize(serializer)
            defined_execution_errors.update(command.errors)

        for property_ in self.properties.values():
            property_.serialize(serializer)
            defined_execution_errors.update(property_.errors)

        for metadata in self.metadata.values():
            metadata.serialize(serializer)
            defined_execution_errors.update(metadata.errors)

        for Error in defined_execution_errors.values():
            Error.serialize(serializer)

        for DataTypeDefinition in self.data_type_definitions.values():
            DataTypeDefinition.serialize(serializer, definition=True)

        serializer.end_element("Feature")

    @typing.override
    def deserialize(self, deserializer: "Deserializer") -> collections.abc.Generator[None, typing.Any, typing.Self]: ...

    @typing.override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Feature):
            return NotImplemented

        return other.fully_qualified_identifier == self.fully_qualified_identifier



import collections.abc
import dataclasses

import typing_extensions as typing

from sila import datetime

from ..data_types.date import Date
from ..data_types.integer import Integer
from ..data_types.real import Real
from ..data_types.time import Time
from ..data_types.timestamp import Timestamp
from ..fdl import Deserializer, Serializer
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", int, float, datetime.date, datetime.time, datetime.datetime)


@dataclasses.dataclass
class MaximalExclusive(Constraint[T]):
    """
    A constraint that enforces an upper exclusive bound on a value.

    This constraint ensures that a given value is strictly less than
    the specified maximum exclusive bound. It can be applied to
    various types, including Integer, Real, Date, Time andTimestamp.
    The constraint raises errors if the input value is greater than
    or equal to the upper bound or is of an incorrect type.

    Args:
      value: The upper exclusive limit for the value. The value must
        be of the same type as this attribute and must be strictly
        less than this limit.
    """

    value: T
    """
    The upper exclusive limit for the value. The value must be of
    the same type as this attribute and must be strictly less
    than this limit.
    """

    async def validate(self, value: T) -> bool:
        if not isinstance(value, type(self.value)):
            raise TypeError(f"Expected value of type '{type(self.value).__name__}', received '{type(value).__name__}'.")

        if value >= self.value:
            raise ValueError(
                f"Value '{value}' must be strictly less than the maximal exclusive limit of '{self.value}'."
            )

        return True

    def serialize(self, serializer: Serializer) -> None:
        if isinstance(self.value, int):
            serializer.write_int("MaximalExclusive", self.value)
        elif isinstance(self.value, float):
            serializer.write_float("MaximalExclusive", self.value)
        elif isinstance(self.value, datetime.date):
            serializer.write_date("MaximalExclusive", self.value)
        elif isinstance(self.value, datetime.time):
            serializer.write_time("MaximalExclusive", self.value)
        elif isinstance(self.value, datetime.datetime):
            serializer.write_datetime("MaximalExclusive", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (Integer, Real, Date, Time, Timestamp)):
            raise ValueError(
                f"Expected constraint's data type to be 'Integer', 'Real', 'Date', 'Time' or 'Timestamp', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="MaximalExclusive")
        if issubclass(data_type, Integer):
            value = yield from deserializer.read_int()
        elif issubclass(data_type, Real):
            value = yield from deserializer.read_float()
        elif issubclass(data_type, Date):
            value = yield from deserializer.read_date()
        elif issubclass(data_type, Time):
            value = yield from deserializer.read_time()
        elif issubclass(data_type, Timestamp):
            value = yield from deserializer.read_datetime()
        yield from deserializer.read_end_element(name="MaximalExclusive")

        return cls(typing.cast(T, value))



import collections.abc
import dataclasses
import enum

import typing_extensions as typing

from ..data_types.integer import Integer
from ..data_types.real import Real
from ..fdl import Characters, Deserializer, EndElement, Serializer, StartElement
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", int, float)


class SIUnit(str, enum.Enum):
    """
    Enumeration of SI units used in measurement.

    This enumeration defines the standard SI units that can be used
    for measurements, including both base and derived units.
    """

    DIMENSIONLESS = "Dimensionless"
    """Represents a dimensionless quantity."""

    METER = "Meter"
    """Represents the meter (m) unit of length."""

    KILOGRAM = "Kilogram"
    """Represents the kilogram (kg) unit of mass."""

    SECOND = "Second"
    """Represents the second (s) unit of time."""

    AMPERE = "Ampere"
    """Represents the ampere (A) unit of electric current."""

    KELVIN = "Kelvin"
    """Represents the kelvin (K) unit of temperature."""

    MOLE = "Mole"
    """Represents the mole (mol) unit of amount of substance."""

    CANDELA = "Candela"
    """Represents the candela (cd) unit of luminous intensity."""


class UnitComponent(typing.NamedTuple):
    """
    Represents a component of a unit in terms of its base unit and
    exponent.

    This named tuple defines a base unit and its exponent, allowing
    for the representation of complex units through their components.

    Args:
      unit: The base unit represented as a string (e.g., 'Meter').
      exponent: The exponent to which the base unit is raised,
    """

    unit: typing.Literal[
        "Dimensionless",
        "Meter",
        "Kilogram",
        "Second",
        "Ampere",
        "Kelvin",
        "Mole",
        "Candela",
    ]
    """
    The base unit represented as a string (e.g., 'Meter').
    """

    exponent: int = 1
    """
    The exponent to which the base unit is raised.
    """


@dataclasses.dataclass
class Unit(Constraint[T]):
    """
    A constraint that defines a unit of measurement with its label
    and components.

    This class allows for the definition of units of measurement
    including their label, components, scaling factor, and offset. It
    ensures that the label does not exceed a specified length.

    Args:
      label: A string representing the label of the unit.
      components: A sequence of `UnitComponent` defining the
        composition of the unit.
      factor: A scaling factor for the unit.
      offset: An offset value for the unit.

    Raises:
      ValueError: If the length of the label exceeds 255 characters.
    """

    Component: typing.ClassVar[type[UnitComponent]] = UnitComponent
    """
    Represents a component of a unit in terms of its base unit and
    exponent.
    """

    SI: typing.ClassVar[type[SIUnit]] = SIUnit
    """
    Enumeration of SI units used in measurement.
    """

    label: str
    """
    A string representing the label of the unit.
    """

    components: collections.abc.Sequence[UnitComponent]
    """
    A sequence of `UnitComponent` defining the composition of the
    unit.
    """

    factor: float = 1
    """
    A scaling factor for the unit (default is 1).
    """

    offset: float = 0
    """
    An offset value for the unit (default is 0).
    """

    def __post_init__(self):
        if len(self.label) > 255:
            raise ValueError("The length of the label must not exceed 255 characters.")

        self.components = [
            UnitComponent(
                component.unit.value if isinstance(component.unit, SIUnit) else component.unit, component.exponent
            )
            for component in self.components
        ]

    async def validate(self, value: T) -> bool:
        if not isinstance(value, (int, float)):
            raise TypeError(f"Expected value of type 'int' or 'float', received '{type(value).__name__}'.")

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.start_element("Unit")
        serializer.write_str("Label", self.label)
        serializer.write_float("Factor", self.factor)
        serializer.write_float("Offset", self.offset)
        for component in self.components:
            serializer.start_element("UnitComponent")
            serializer.write_str("SIUnit", component.unit)
            serializer.write_int("Exponent", component.exponent)
            serializer.end_element("UnitComponent")

        serializer.end_element("Unit")

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (Integer, Real)):
            raise ValueError(
                f"Expected constraint's data type to be 'Integer' or 'Real', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="Unit")

        # Label
        yield from deserializer.read_start_element(name="Label")
        label = yield from deserializer.read_str()
        yield from deserializer.read_end_element(name="Label")

        # Factor
        yield from deserializer.read_start_element(name="Factor")
        factor = yield from deserializer.read_float()
        yield from deserializer.read_end_element(name="Factor")

        # Offset
        yield from deserializer.read_start_element(name="Offset")
        offset = yield from deserializer.read_float()
        yield from deserializer.read_end_element(name="Offset")

        components: list[UnitComponent] = []
        while True:
            token = yield

            if isinstance(token, StartElement):
                if token.name == "UnitComponent":
                    # SIUnit
                    yield from deserializer.read_start_element("SIUnit")
                    unit = yield from deserializer.read_str()
                    try:
                        unit = Unit.SI(unit).value
                    except ValueError:
                        raise ValueError(f"Expected a valid 'SIUnit' value, received '{unit}'.") from None
                    yield from deserializer.read_end_element("SIUnit")

                    # Exponent
                    yield from deserializer.read_start_element("Exponent")
                    exponent = yield from deserializer.read_int()
                    yield from deserializer.read_end_element("Exponent")

                    components.append(UnitComponent(unit, exponent))
                else:
                    raise ValueError(
                        f"Expected start element with name 'UnitComponent', received start element with name '{token.name}'."
                    )

            elif isinstance(token, EndElement):
                if token.name == "UnitComponent":
                    continue
                else:
                    break  # pragma: no cover

            elif isinstance(token, Characters):
                raise ValueError(
                    f"Expected start element with name 'UnitComponent', received characters '{token.value}'."
                )

        if not components:
            raise ValueError("Expected at least one 'UnitComponent' element inside the 'Unit' element.")

        return cls(label, components, factor, offset)



import dataclasses

import typing_extensions as typing

from ..protobuf import Reader, WireType, Writer
from .data_type import BasicType

if typing.TYPE_CHECKING:
    from ..common import Context


@dataclasses.dataclass
class Binary(BasicType[bytes]):
    """
    Represents arbitrary binary data of any size such as images, flat files, etc.

    If the SiLA binary type is used for character data, e.g. plain
    text, XML or JSON, the character encoding must be UTF-8. It is
    recommended to specify constraints, e.g. a content-type or schema
    constraint for the SiLA binary type in order to make the binary
    content type safe.

    Attributes:
      value: The encapsulated `bytes` value. Defaults to an empty
        bytes.
      binary_transfer_uuid: When the value is too large, the binary
        is transferred via the SiLA binary transfer feature,
        identified with this uuid.
    """

    value: bytes = b""
    binary_transfer_uuid: typing.Optional[str] = None

    @typing.override
    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[bytes] = None, /) -> typing.Self:
        if value is None:
            return await cls().validate()

        try:
            size = len(value)
        except TypeError:
            size = 0

        if size <= 2**21:
            return await cls(value=value).validate()

        binary_transfer_uuid = await context.binary_transfer_handler.set_binary(value)

        return await cls(binary_transfer_uuid=binary_transfer_uuid).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> bytes:
        await self.validate()

        if self.binary_transfer_uuid is not None:
            return await context.binary_transfer_handler.get_binary(self.binary_transfer_uuid)

        return self.value

    @typing.override
    async def validate(self) -> typing.Self:
        if not isinstance(self.value, bytes):
            msg = f"Expected value of type 'bytes', received '{type(self.value).__name__}'."
            raise TypeError(msg)

        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                message.value = reader.read_bytes()
                message.binary_transfer_uuid = None
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message.value = b""
                message.binary_transfer_uuid = reader.read_string()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.binary_transfer_uuid:
            writer.write_uint32(18).write_string(self.binary_transfer_uuid)
        else:
            writer.write_uint32(10).write_bytes(self.value)

        if number:
            writer.ldelim()

        return writer.finish()



import functools
import re

import typing_extensions as typing

originator = r"(?P<originator>[a-z][a-z\.]*)"
category = r"(?P<category>[a-z][a-z\.]*)"
identifier = r"[A-Z][a-zA-Z0-9]*"
version = r"v(?P<version>\d+)"


class FeatureIdentifier(str):
    """
    Represents a unique identifier for a feature.

    Attributes:
      originator: Organization that created and owns the feature.
      category: Domain or group of the feature.
      feature: Unique feature name.
      version: Major version number.

    Raises:
      ValueError: If the input string does not match the required format.

    Examples:
      >>> identifier = FeatureIdentifier("org.silastandard/core/SiLAService/v1")
    """

    _registry: typing.ClassVar[set[type[typing.Self]]] = set()
    _pattern: typing.ClassVar[str] = rf"{originator}/{category}/(?P<feature>{identifier})/{version}"
    _regex: typing.ClassVar[re.Pattern] = re.compile(_pattern)

    _data: dict[str, typing.Any]

    @typing.override
    def __new__(cls, value: str) -> typing.Self:
        if isinstance(value, cls):
            instance = super().__new__(type(value), value)
            instance._data = value._data

            return instance

        registry = {cls} | cls._registry if cls is FeatureIdentifier else {cls}
        for identifier in registry:
            if match := identifier._regex.fullmatch(value):
                values = match.groupdict()

                if "version" in values:
                    values["version"] = int(values["version"])

                instance = super().__new__(identifier, value)
                instance._data = values

                return instance

        msg = f"Invalid format for {cls.__name__} {value!r}."
        raise ValueError(msg)

    @typing.override
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls._registry.add(cls)

    @classmethod
    def create(cls, originator: str, category: str, feature: str, version: int) -> typing.Self:
        """
        Create a new feature identifier from the given components.

        Args:
          originator: The organization who created and owns the feature.
          category: Group features into domains of application.
          feature: Unique feature identifier.
          version: The major feature version.

        Returns:
          The created feature identifier.
        """

        identifier = super().__new__(cls, f"{originator}/{category}/{feature}/v{version}")
        identifier._data = {
            "originator": originator,
            "category": category,
            "feature": feature,
            "version": version,
        }

        return identifier

    @property
    def originator(self) -> str:
        """The organization who created and owns the feature."""

        return self._data["originator"]

    @property
    def category(self) -> str:
        """Group features into domains of application."""

        return self._data["category"]

    @property
    def feature(self) -> str:
        """Unique feature identifier."""

        return self._data["feature"]

    @property
    def version(self) -> int:
        """The major feature version."""

        return self._data["version"]

    @functools.cached_property
    def feature_identifier(self) -> "FeatureIdentifier":
        """The fully qualified feature identifier."""

        return FeatureIdentifier.create(self.originator, self.category, self.feature, self.version)

    @typing.override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, str):
            return NotImplemented

        return self.lower() == other.lower()

    @typing.override
    def __ne__(self, other: object) -> bool:
        if not isinstance(other, str):
            return NotImplemented

        return self.lower() != other.lower()

    @typing.override
    def __hash__(self):
        return super().lower().__hash__()

    @typing.override
    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"value={super().__str__()!r}, "
            f"{', '.join(f'{key}={value!r}' for key, value in self._data.items())}"
            ")"
        )



import collections.abc

import typing_extensions as typing


class ConversionError(Exception):
    """
    Raised when converting operations encounter invalid data.

    Args:
      message: An explanation of why the conversion error occurred.
      path: The path to the element where the error occurred.
    """

    def __init__(self, message: str, path: typing.Optional[collections.abc.Sequence[str]] = None):
        super().__init__(message, list(path) if path is not None else [])

    @property
    def message(self) -> str:
        """An explanation of why the conversion error occurred."""

        return self.args[0]

    @property
    def path(self) -> collections.abc.Sequence[str]:
        """The path to the element where the error occurred."""

        return self.args[1]

    def __str__(self) -> str:
        return self.args[0]



import dataclasses
import inspect
import logging

import grpc
import grpc.aio
import typing_extensions as typing

from .. import framework
from ..framework import (
    ConversionError,
    DecodeError,
    DefinedExecutionError,
    Feature,
    MetadataIdentifier,
    Native,
    NoMetadataAllowed,
    Server,
    SiLAError,
    UndefinedExecutionError,
    ValidationError,
)


@dataclasses.dataclass
class UnobservableCommand(framework.UnobservableCommand):
    """Any command for which observing the progress of execution is not possible or does not make sense."""

    function: typing.Callable = dataclasses.field(repr=False, default=lambda **_: ...)
    """The implementation which is executed by the RPC handler."""

    @property
    def logger(self) -> logging.Logger:
        """A python logger instance."""

        return logging.getLogger(__name__)

    @typing.override
    async def execute(self, request: bytes = b"", metadata: typing.Optional[dict[str, bytes]] = None) -> bytes:
        assert self.feature is not None and isinstance(self.feature.context, Server)

        try:
            # Metadata
            if (
                metadata
                and self.fully_qualified_identifier.feature_identifier == "org.silastandard/core/SiLAService/v1"
            ):
                msg = "No metadata allowed for the SiLA Service feature."
                raise NoMetadataAllowed(msg)

            parsed_metadata = dict[MetadataIdentifier, Native]()
            for interceptor in self.feature.context.get_metadata_by_affect(self.fully_qualified_identifier):
                metadatum = await interceptor.from_buffer(self, metadata)
                parsed_metadata.update(await metadatum.intercept(self))

            # Parameters
            try:
                parameters = await self.feature.context.protobuf.decode(
                    f"{self.feature.rpc_package}.{self.identifier}_Parameters", request
                )
            except DecodeError as error:
                raise ValidationError(
                    parameter=f"{self.fully_qualified_identifier}/Parameter/{error.path[0]}", message=error.message
                ) from error
            except ConversionError as error:
                raise ValidationError(
                    parameter=f"{self.fully_qualified_identifier}/Parameter/{error.path[0]}", message=error.message
                ) from error

            # Execute
            responses = self.function(**parameters, metadata=parsed_metadata)

            if inspect.isawaitable(responses):
                responses = await responses

            return await self.feature.context.protobuf.encode(
                f"{self.feature.rpc_package}.{self.identifier}_Responses", responses
            )
        except SiLAError as error:
            if isinstance(error, DefinedExecutionError) and error._identifier is None:
                error = error.with_feature(self.feature.fully_qualified_identifier)

            raise error
        except Exception as error:
            self.logger.exception(error)
            raise UndefinedExecutionError(str(error)) from error

    async def execute_rpc_handler(self, request: bytes, context: grpc.aio.ServicerContext) -> bytes:
        """
        Handle the gRPC call to execute the command.

        Args:
          request: The request payload in protobuf ecoding.
          context: The gRPC call context.

        Returns:
          The response payload in protobuf ecoding.
        """

        try:
            metadata: dict[str, bytes] = {
                key: value
                for key, value in context.invocation_metadata() or ()
                if key.startswith("sila-") and key.endswith("-bin") and isinstance(value, bytes)
            }

            return await self.execute(request, metadata)
        except SiLAError as error:
            raise await error.to_rpc_error(context) from None

    @typing.override
    def add_to_feature(self, feature: "Feature") -> typing.Self:
        super().add_to_feature(feature)

        feature.context.protobuf.register_message(
            name=self.identifier + "_Parameters", message=self.parameters, package=feature.rpc_package
        )
        feature.context.protobuf.register_message(
            name=self.identifier + "_Responses", message=self.responses, package=feature.rpc_package
        )
        feature.context.protobuf.register_service(
            feature.identifier,
            {self.identifier: grpc.unary_unary_rpc_method_handler(self.execute_rpc_handler)},
            package=feature.rpc_package,
        )

        return self



from .observable_property import ObservableProperty
from .property import Property
from .unobservable_property import UnobservableProperty

__all__ = [
    "ObservableProperty",
    "Property",
    "UnobservableProperty",
]



import enum

import typing_extensions as typing

from ..errors import SiLAError
from ..protobuf import Reader, WireType, Writer


class BinaryTransferErrorType(enum.IntEnum):
    """Type of errors that occur during binary transfer operations."""

    INVALID_BINARY_TRANSFER_UUID = 0
    """
    Is issued if the provided binary transfer UUID is invalid or not
    recognized.
    """

    BINARY_UPLOAD_FAILED = 1
    """
    Is issued when the upload of binary data has failed.
    """

    BINARY_DOWNLOAD_FAILED = 2
    """
    Is issued when the download of binary data has failed.
    """


class BinaryTransferError(SiLAError):
    """
    Error that occurs during binary transfer operations.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
      error_type: The type of error that occurred.
    """

    Type: typing.ClassVar[type[BinaryTransferErrorType]] = BinaryTransferErrorType

    def __init__(self, message: str, error_type: BinaryTransferErrorType) -> None:
        super().__init__(message, error_type)

    @property
    def error_type(self) -> BinaryTransferErrorType:
        """The type of error that occurred."""

        return self.args[1]

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = ""
        error_type = BinaryTransferErrorType.INVALID_BINARY_TRANSFER_UUID
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                error_type = BinaryTransferErrorType(reader.read_int32())
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message = reader.read_string()
            else:
                reader.skip_type(tag & 7)

        return cls(message, error_type)

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.error_type:
            writer.write_uint32(8).write_int32(self.error_type)
        if self.message:
            writer.write_uint32(18).write_string(self.message)

        if number:
            writer.ldelim()

        return writer.finish()


class InvalidBinaryTransferUUID(BinaryTransferError):
    """
    Raised when the provided identifier is invalid or not recognized.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str, *args) -> None:
        super().__init__(message, BinaryTransferErrorType.INVALID_BINARY_TRANSFER_UUID)


class BinaryUploadFailed(BinaryTransferError):
    """
    Raised when the upload of binary data has failed.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str, *args) -> None:
        super().__init__(message, BinaryTransferErrorType.BINARY_UPLOAD_FAILED)


class BinaryDownloadFailed(BinaryTransferError):
    """
    Raised when the download of binary data has failed.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str, *args) -> None:
        super().__init__(message, BinaryTransferErrorType.BINARY_DOWNLOAD_FAILED)



import dataclasses

import typing_extensions as typing

from .command import Command


@dataclasses.dataclass
class UnobservableCommand(Command):
    """Any command for which observing the progress of execution is not possible or does not make sense."""

    observable: bool = dataclasses.field(init=False, repr=False, default=False)

    async def execute(self, request: bytes = b"", metadata: typing.Optional[dict] = None) -> bytes:
        """
        Execute the command.

        Args:
          request: Input parameters passed to the command.
          metadata: Additional metadata sent from client to server.

        Returns:
          The resulting responses of the command execution.

        Raises:
          NoMetadataAllowed: If providing metadata is not allowed.
          InvalidMetadata: If metadata is missing or invalid.
          CommandExecutionNotAccepted: If the server does not allow
            further command executions.
          ValidationError: If command parameters are missing or invalid.
          DefinedExecutionError: If command execution results in a
            defined execution error
          UnefinedExecutionError: If command execution results in an
            undefined execution error.
        """

        raise NotImplementedError



import collections.abc

import typing_extensions as typing


class ParseError(Exception):
    """
    Raised when XML parsing operations encounter unexpected or invalid data.

    Args:
      message: An explanation of why the parsing error occurred.
      path: The path to the element where the error occurred.
      line: The line number in which the current error occurred.
      column: The column number in which the current error occurred.
    """

    def __init__(
        self,
        message: str,
        /,
        *,
        path: typing.Optional[collections.abc.Sequence[str]] = None,
        line: int = -1,
        column: int = -1,
    ):
        super().__init__(message, list(path) if path is not None else [], line, column)

    @property
    def message(self) -> str:
        """An explanation of why the conversion error occurred."""

        return self.args[0]

    @property
    def path(self) -> collections.abc.Sequence[str]:
        """The path to the element where the error occurred."""

        return self.args[1]

    @property
    def line(self) -> int:
        """The line number in which the current error occurred."""

        return self.args[2]

    @property
    def column(self) -> int:
        """The column number in which the current error occurred."""

        return self.args[3]

    def __str__(self) -> str:
        return self.args[0]



import enum
import re

import typing_extensions as typing


class VersionLevel(enum.IntEnum):
    """Different levels of specificity of a version. A version is written as "major.minor.patch_label"."""

    MAJOR = 1
    MINOR = 2
    PATCH = 3
    LABEL = 4


def check_version(value: str, /, required: VersionLevel, optional: typing.Optional[VersionLevel] = None) -> str:
    """
    Validate whether the given value is a semantic version.

    Args:
      value: The input value to validate.
      required: The minimum level of detail required.
      optional: The minimum level of detail allowed.

    Returns:
      The input value, allowing for method chaining.

    Raises:
      ValueError: If the provided value does not follow the expected
        schema.
    """

    optional = optional or required
    if required.value > optional.value:
        msg = "Optional level can not be less detailed than required level."
        raise ValueError(msg)

    if not isinstance(value, str):
        msg = f"Version must be a string of unicode characters, received '{value}'."
        raise ValueError(msg)

    version = value
    if optional == VersionLevel.LABEL:
        version, _, label = value.partition("_")

        if required == VersionLevel.LABEL and not label:
            msg = f"Version must contain a label after an underscore, received '{value}'."
            raise ValueError(msg)

        if not re.fullmatch(r"[a-zA-Z0-9\_]*", label):
            msg = f"Version label may only contain letters, digits and underscores, received '{value}'."
            raise ValueError(msg)

    parts = version.split(".")
    if len(parts) < min(required.value, VersionLevel.PATCH):
        msg = f"Version must contain at least {required.value} parts separated by dots, received '{value}'."
        raise ValueError(msg)

    if len(parts) > min(optional.value, VersionLevel.PATCH):
        msg = f"Version must contain at most {optional.value} parts separated by dots, received '{value}'."
        raise ValueError(msg)

    for item in parts:
        if not item.isdigit():
            msg = f"Version parts must represent a numeric value, received '{value}'."
            raise ValueError(msg)

    return value



import dataclasses

import typing_extensions as typing

from ..protobuf import Message, Reader, WireType, Writer


@dataclasses.dataclass
class MetadataResponse(Message):
    """
    Response with calls affected by the requested metadata.

    Attributes:
      affected_calls: The list of calls affected by the requested
        metadata.
    """

    affected_calls: list[str] = dataclasses.field(default_factory=list)

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                value = reader.read_string()
                message.affected_calls.append(value)
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        for affected_call in self.affected_calls:
            writer.write_uint32(10).write_string(affected_call)

        if number:
            writer.ldelim()

        return writer.finish()



import contextlib
import struct

import typing_extensions as typing

from .decode_error import DecodeError
from .wire_type import WireType


class Reader:
    """
    A data source from which to decode a Protocol Buffer message.

    Args:
      buffer: A bytes object containing the data to be read.
    """

    def __init__(self, buffer: bytes) -> None:
        self.buf = bytearray(buffer)
        self.len = len(buffer)
        self.cur = 0

    @property
    def buffer(self) -> bytearray:
        """
        The internal buffer that holds the provided byte data.
        """

        return self.buf

    @property
    def length(self) -> int:
        """
        The total length of the byte buffer.
        """

        return self.len

    @property
    def cursor(self) -> int:
        """
        The current read position within the buffer.
        """

        return self.cur

    def read_varint(self) -> int:
        """
        Decode a variable-length integer encoded in the buffer.

        Returns:
          The decoded variable-length integer.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        value = 0
        shift = 0

        while True:
            # Check if we've reached the end of the buffer
            if self.cur >= self.len:
                raise DecodeError(
                    "Buffer does not contain enough bytes to decode a variable-length integer.", offset=self.cur
                )

            byte = self.buf[self.cur]
            self.cur += 1
            value |= (byte & 0x7F) << shift

            # If the MSB is 0, this is the last byte of the varint
            if byte & 0x80 == 0:
                return value

            shift += 7

            # Check for overflow; varints are limited to 64 bits
            if shift >= 64:
                raise DecodeError("Variable-length integer is too large.", offset=self.cur)

    def read_uint32(self) -> int:
        """
        Read a variable-length unsigned 32-bit integer from the current
        buffer position.

        Returns:
          The decoded unsigned 32-bit integer.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        return self.read_varint()

    def read_int32(self) -> int:
        """
        Read a variable-length signed 32-bit integer from the current
        buffer position.

        Returns:
          The decoded signed 32-bit integer.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        value = self.read_varint()

        return value if value < (1 << 63) else value - (1 << 64)

    def read_uint64(self) -> int:
        """
        Read a variable-length unsigned 64-bit integer from the current
        buffer position.

        Returns:
          The decoded unsigned 64-bit integer.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        return self.read_varint()

    def read_int64(self) -> int:
        """
        Read a variable-length signed 64-bit integer from the current
        buffer position.

        Returns:
          The decoded signed 64-bit integer.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        value = self.read_varint()

        return value if value < (1 << 63) else value - (1 << 64)

    def read_double(self) -> float:
        """
        Read a 64-bit double-precision floating-point number from the
        current buffer position.

        Returns:
          The decoded double-precision floating-point number.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        parser = struct.Struct("<d")
        try:
            value = float(parser.unpack_from(self.buf, self.cur)[0])
            self.cur += 8
        except struct.error as error:
            raise DecodeError(str(error), offset=self.cur)

        return value

    def read_bool(self) -> bool:
        """
        Read a boolean value from the current buffer position.

        Returns:
          The decoded boolean value.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        return self.read_uint32() == 1

    def read_bytes(self) -> bytes:
        """
        Read a length-prefixed byte sequence from the current buffer
        position.

        Returns:
          The extracted byte sequence.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        # Read the length of the byte sequence
        length = self.read_uint32()
        end = self.cur + length

        if end > self.len:
            raise DecodeError(
                f"Attempted to read {length} bytes at position {self.cur}, but the buffer only has "
                f"{self.len - self.cur} remaining bytes. Cannot read past the end of the buffer.",
                offset=self.cur,
            )

        # Extract the byte sequence
        result = self.buf[self.cur : end]
        self.cur = end

        return bytes(result)

    def read_string(self) -> str:
        """
        Read a length-prefixed string sequence from the current buffer
        position.

        Returns:
          The extracted string sequence.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        return self.read_bytes().decode("utf-8")

    def skip(self, length: typing.Optional[int] = None) -> "Reader":
        """
        Advance the reader's position in the buffer by a specified number
        of bytes.

        Args:
          length: The number of bytes to skip. If `None`, skips a varint.

        Returns:
          The current instance, allowing method chaining.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        if length is not None:
            if self.cur + length > self.len:
                raise DecodeError(
                    f"Attempted to skip {length} bytes from position {self.cur}, but the buffer only has "
                    f"{self.len - self.cur} remaining bytes. Cannot read past the end of the buffer.",
                    offset=self.cur,
                )

            self.cur += length
        else:
            self.read_uint32()

        return self

    def skip_type(self, wire_type: int) -> "Reader":
        """
        Skip over a protobuf field based on its wire type.

        Args:
          wire_type: The protobuf wire type to skip.

        Returns:
          The current instance, allowing method chaining.

        Raises:
          DecodeError: If the buffer does not contain enough bytes or is
            too large.
        """

        offset = self.cur
        if wire_type == WireType.VARINT:
            self.skip()
        elif wire_type == WireType.I64:
            self.skip(8)
        elif wire_type == WireType.LEN:
            self.skip(self.read_uint32())
        elif wire_type == WireType.I32:
            self.skip(4)
        else:
            value = str(wire_type)
            with contextlib.suppress(ValueError):
                value = WireType(wire_type).name

            raise DecodeError(f"Invalid wire type '{value}' at offset '{offset}'.", offset)

        return self

    def expect_type(self, tag: int, wire_type: WireType) -> None:
        if tag & 7 != wire_type:
            msg = f"Expected wire type '{wire_type.name}', received '{WireType(tag & 7).name}'."

            raise DecodeError(msg, self.cur)



import dataclasses

import typing_extensions as typing

from ..protobuf import Message, Reader, WireType, Writer


@dataclasses.dataclass
class UploadChunkRequest(Message):
    """
    Upload a specific chunk of binary data.

    Attributes:
      binary_transfer_uuid: A unique identifier (UUID) for the binary
        transfer session to which this chunk is being uploaded.
      chunk_index: The sequential index of the chunk within the
        binary transfer session. This helps the server to correctly
        assemble the chunks in order.
      payload: The actual binary data chunk being uploaded to the
        server.
    """

    binary_transfer_uuid: str = ""
    chunk_index: int = 0
    payload: bytes = b""

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                message.binary_transfer_uuid = reader.read_string()
            elif field_number == 2:
                reader.expect_type(tag, WireType.VARINT)
                message.chunk_index = reader.read_uint32()
            elif field_number == 3:
                reader.expect_type(tag, WireType.LEN)
                message.payload = reader.read_bytes()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.binary_transfer_uuid:
            writer.write_uint32(10).write_string(self.binary_transfer_uuid)
        if self.chunk_index:
            writer.write_uint32(16).write_uint32(self.chunk_index)
        if self.payload:
            writer.write_uint32(26).write_bytes(self.payload)

        if number:
            writer.ldelim()

        return writer.finish()



import re

import typing_extensions as typing

from .feature_identifier import FeatureIdentifier, identifier


class ErrorIdentifier(FeatureIdentifier):
    """
    Uniquely identifies a defined execution error among all potentially existing errors.

    Attributes:
      error: Unique defined execution error identifier.

    Examples:
      >>> identifier = FeatureIdentifier(
      ...     "org.silastandard/core/SiLAService/v1/DefinedExecutionError/UnimplementedFeature"
      ... )
    """

    _pattern: typing.ClassVar[str] = f"{FeatureIdentifier._pattern}/DefinedExecutionError/(?P<error>{identifier})"
    _regex: typing.ClassVar[re.Pattern] = re.compile(_pattern, re.IGNORECASE)

    @typing.override
    @classmethod
    def create(cls, originator: str, category: str, feature: str, version: int, error: str) -> typing.Self:
        """
        Create a new defined execution error identifier from the given components.

        Args:
          originator: The organization who created and owns the feature.
          category: Group features into domains of application.
          feature: Unique feature identifier.
          version: The major feature version.
          error: Unique defined execution error identifier.

        Returns:
          The created defined execution error identifier.
        """

        identifier = super().__new__(cls, f"{originator}/{category}/{feature}/v{version}/DefinedExecutionError/{error}")
        identifier._data = {
            "originator": originator,
            "category": category,
            "feature": feature,
            "version": version,
            "error": error,
        }

        return identifier

    @property
    def error(self) -> str:
        """Unique defined execution error identifier."""

        return self._data["error"]



import collections.abc
import dataclasses

import typing_extensions as typing

from .command import Command

if typing.TYPE_CHECKING:
    from ..data_types import Element
    from .command_confirmation import CommandConfirmation
    from .command_execution_info import CommandExecutionInfo


@dataclasses.dataclass
class ObservableCommand(Command):
    """Any command for which observing the progress of execution is possible or does make sense."""

    observable: bool = dataclasses.field(init=False, repr=False, default=True)
    """Whether the handler returns an observable stream or just a single value."""

    intermediate_responses: dict[str, "Element"] = dataclasses.field(default_factory=dict)
    """An intermediate response of the command execution."""

    async def initiate(self, request: bytes = b"", metadata: typing.Optional[dict] = None) -> "CommandConfirmation":
        """
        Initiate the execution of  the command.

        Args:
          request: Input parameters passed to the command.
          metadata: Additional metadata sent from client to server.

        Returns:
          Confirmation that the command execution has been accepted.

        Raises:
          NoMetadataAllowed: If providing metadata is not allowed.
          InvalidMetadata: If metadata is missing or invalid.
          CommandExecutionNotAccepted: If the server does not allow
            further command executions.
          ValidationError: If command parameters are missing or invalid.
          DefinedExecutionError: If metadata handling results in a
            defined execution error
        """

        raise NotImplementedError

    def subscribe_status(self, command_execution_uuid: str) -> collections.abc.AsyncIterator["CommandExecutionInfo"]:
        """
        Subscribe to status changes of the command execution.

        Args:
          command_execution_uuid: The unique identifier of the command
            execution.

        Yields:
          The current status of the command execution.

        Raises:
          InvalidCommandExecutionUUID: If the given identifier is invalid
            or not recognized.
        """

        raise NotImplementedError

    def subscribe_intermediate(self, command_execution_uuid: str) -> collections.abc.AsyncIterator[bytes]:
        """
        Subscribe to intermediate responses of the command execution.

        Args:
          command_execution_uuid: The unique identifier of the command
            execution.

        Yields:
          The current intermediate responses of the command execution.

        Raises:
          InvalidCommandExecutionUUID: If the given identifier is invalid
            or not recognized.
        """

        raise NotImplementedError

    async def get_result(self, command_execution_uuid: str) -> bytes:
        """
        Get the responses of the command execution.

        Args:
          command_execution_uuid: The unique identifier of the command
            execution.

        Returns:
          The resulting responses of the command execution.

        Raises:
          InvalidCommandExecutionUUID: If the given identifier is invalid
            or not recognized.
          CommandExecutionNotFinished: If the command execution has not
            been finished yet.
          DefinedExecutionError: If command execution resulted in a
            defined execution error
          UnefinedExecutionError: If command execution resulted in an
            undefined execution error.
        """

        raise NotImplementedError



import dataclasses

import typing_extensions as typing

from ..protobuf import Message, Reader, Writer


@dataclasses.dataclass
class CancelRequest(Message):
    """Message to cancel an ongoing subscription."""

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()
            writer.ldelim()

        return writer.finish()



import collections.abc
import contextlib

import typing_extensions as typing

from ..fdl import Serializable
from ..identifiers import ErrorIdentifier, FeatureIdentifier
from ..protobuf import Reader, WireType, Writer
from ..validators import check_display_name, check_identifier
from .sila_error import SiLAError

if typing.TYPE_CHECKING:
    from ..fdl import Deserializer, Serializer


class DefinedExecutionError(SiLAError, Serializable):
    """
    Expected error that occurs during command execution.

    Attributes:
      identifier: Uniquely identifies the defined execution error
        within the scope of the same feature.
      display_name: Human readable name of the defined execution
        error.
      description: Describes the use and purpose of the defined
        execution error.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    identifier: typing.ClassVar[str] = ""
    display_name: typing.ClassVar[str] = ""
    description: typing.ClassVar[str] = ""

    def __init__(self, message: str = "") -> None:
        super().__init__(message or self.description)

        self._identifier: typing.Optional[ErrorIdentifier] = None

    @property
    def fully_qualified_identifier(self) -> ErrorIdentifier:
        """Uniquely identifies the defined execution error."""

        if self._identifier is None:
            msg = (
                f"Unable to get fully qualified identifier for DefinedExecutionError "
                f"'{self.identifier}' without feature association."
            )
            raise RuntimeError(msg)

        return self._identifier

    def with_feature(self, feature_identifier: str, /) -> typing.Self:
        """
        Assign the defined execution error to the given feature.

        Args:
          feature_identifier: The fully qualified feature identifier this
            error is assigned to.

        Returns:
          The Error instance, allowing for method chaining.
        """

        feature_identifier = FeatureIdentifier(feature_identifier)
        self._identifier = ErrorIdentifier.create(**feature_identifier.feature_identifier._data, error=self.identifier)

        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = ""
        identifier = ""
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                identifier = reader.read_string()
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message = reader.read_string()
            else:
                reader.skip_type(tag & 7)

        error = cls(message)
        with contextlib.suppress(ValueError):
            error._identifier = ErrorIdentifier(identifier)

        return error

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        writer.write_uint32(18).fork()

        if self._identifier:
            writer.write_uint32(10).write_string(self._identifier)
        if self.message:
            writer.write_uint32(18).write_string(self.message)

        writer.ldelim()

        if number:
            writer.ldelim()

        return writer.finish()

    @classmethod
    def create(
        cls,
        identifier: str,
        display_name: str,
        description: str = "",
    ) -> type[typing.Self]:
        """
        Create a new SiLA `DefinedExecutionError` class.

        Args:
          identifier: Uniquely identifies the defined execution error
            within the scope of the same feature.
          display_name: Human readable name of the defined execution
            error.
          description: Describes the use and purpose of the defined
            execution error.

        Returns:
          A new `DefinedExecutionError` class with the specified info.
        """

        check_identifier(identifier)
        check_display_name(display_name)

        return typing.cast(
            type[typing.Self],
            type(
                identifier or cls.__name__,
                (cls,),
                {
                    "__doc__": description,
                    "identifier": identifier,
                    "display_name": display_name,
                    "description": description,
                },
            ),
        )

    @typing.override
    @classmethod
    def serialize(cls, serializer: "Serializer") -> None:
        serializer.start_element("DefinedExecutionError")
        serializer.write_str("Identifier", cls.identifier)
        serializer.write_str("DisplayName", cls.display_name)
        serializer.write_str("Description", cls.description)
        serializer.end_element("DefinedExecutionError")

    @typing.override
    @classmethod
    def deserialize(
        cls, deserializer: "Deserializer"
    ) -> collections.abc.Generator[None, typing.Any, type[typing.Self]]: ...



import dataclasses

import typing_extensions as typing

from ..data_types import Duration
from ..protobuf import Message, Reader, WireType, Writer


@dataclasses.dataclass
class GetBinaryInfoResponse(Message):
    """
    The detailed info about a specific binary transfer session.

    Attributes:
      binary_size: The total size of the binary data associated with
        the transfer session, in bytes.
      lifetime_of_binary: The remaining duration for which the binary
        data will be retained on the server before it expires.
    """

    binary_size: int = 0
    lifetime_of_binary: Duration = dataclasses.field(default_factory=Duration)

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                message.binary_size = reader.read_uint64()
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message.lifetime_of_binary = Duration.decode(reader, reader.read_uint32())
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.binary_size:
            writer.write_uint32(8).write_uint64(self.binary_size)
        self.lifetime_of_binary.encode(writer, 2)

        if number:
            writer.ldelim()

        return writer.finish()



import typing_extensions as typing

from ..protobuf import Protobuf
from .context import Context

if typing.TYPE_CHECKING:
    from ..binary_transfer import BinaryTransferHandler
    from ..command import CommandExecution
    from ..identifiers import FeatureIdentifier
    from .feature import Feature


class ContextProxy(Context):
    """Mimics a real context object to enable lazy loading."""

    def __init__(self) -> None:
        self._context: typing.Optional[Context] = None
        self._protobuf: Protobuf = Protobuf(self)
        self._features: dict["FeatureIdentifier", "Feature"] = {}
        self._command_executions: dict[str, "CommandExecution"] = {}

    @property
    def protobuf(self) -> Protobuf:
        """A collection of protobuf messages and services."""

        if self._context:
            return self._context.protobuf

        return self._protobuf

    @property
    def features(self) -> dict["FeatureIdentifier", "Feature"]:
        """A collection of registered features."""

        if self._context:
            return self._context.features

        return self._features

    @property
    def command_executions(self) -> dict[str, "CommandExecution"]:
        """A collection of currently executed commands."""

        if self._context:
            return self._context.command_executions

        return self._command_executions

    @property
    def binary_transfer_handler(self) -> "BinaryTransferHandler":
        """Upload and download large binaries in chunks."""

        if self._context:
            return self._context.binary_transfer_handler

        msg = "Unable to access 'BinaryTransferHandler' on unbound 'Feature'."
        raise RuntimeError(msg)

    @property
    def context(self) -> typing.Optional[Context]:
        """The lazily loaded context that is protected by this proxy."""

        return self._context

    @context.setter
    def context(self, context: Context) -> None:
        context._features |= self._features
        context._protobuf.merge(self._protobuf)
        self._context = context

    @typing.override
    def register_feature(self, feature: "Feature") -> None:
        if self.context is not None:
            self.context.register_feature(feature)
        else:
            self.features[feature.fully_qualified_identifier] = feature

    @typing.override
    def add_command_execution(self, command_execution: "CommandExecution") -> None:
        if self.context is not None:
            self.context.add_command_execution(command_execution)
        else:
            self.command_executions[command_execution.command_execution_uuid] = command_execution

    @typing.override
    def get_command_execution(self, command_execution_uuid: str) -> "CommandExecution":
        try:
            return self.command_executions[command_execution_uuid]
        except KeyError:
            msg = f"Could not find any execution with the given uuid '{command_execution_uuid}'."
            raise ValueError(msg) from None



import typing_extensions as typing

from ..protobuf import Reader, WireType, Writer
from .sila_error import SiLAError


class UndefinedExecutionError(SiLAError):
    """
    Unexpected error that occurs during command execution.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str = "", *args: object) -> None:
        super().__init__(message, *args)

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = ""
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                message = reader.read_string()
            else:
                reader.skip_type(tag & 7)

        return cls(message)

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        writer.write_uint32(26).fork()

        if self.message:
            writer.write_uint32(10).write_string(self.message)

        writer.ldelim()

        if number:
            writer.ldelim()

        return writer.finish()



import collections.abc
import dataclasses

import typing_extensions as typing

from ..data_types.data_type import DataType
from ..fdl import Deserializer, EndElement, Serializer, StartElement
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, DataType, List


@dataclasses.dataclass
class AllowedTypes(Constraint[typing.Any]):
    """
    A constraint that specifies a set of allowed data types for a
    value.

    This class defines a set of permissible data types and raises
    errors if the input value does not match one of the allowed
    types.

    Args:
      values: A sequence of allowed data types that the input value
        must match.
    """

    values: collections.abc.Sequence[type["DataType"]]
    """
    A sequence of allowed data types that the input value must match.
    """

    async def validate(self, value: typing.Any) -> bool:
        # TODO: Validate value is one of the given allowed_types

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.start_element("AllowedTypes")
        for value in self.values:
            value.serialize(serializer)
        serializer.end_element("AllowedTypes")

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        yield from deserializer.read_start_element(name="AllowedTypes")

        data_types: list[type[DataType]] = []
        while True:
            token = yield from deserializer.peek()

            if token == StartElement("DataType"):
                inner_data_type = yield from deserializer.read(DataType.deserialize(deserializer))
                data_types.append(inner_data_type)

            elif token == EndElement("AllowedTypes"):
                break

        yield from deserializer.read_end_element(name="AllowedTypes")

        return cls(data_types)



import collections.abc
import dataclasses

import typing_extensions as typing

from sila import datetime

from ..data_types.date import Date
from ..data_types.integer import Integer
from ..data_types.real import Real
from ..data_types.time import Time
from ..data_types.timestamp import Timestamp
from ..fdl import Deserializer, Serializer
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", int, float, datetime.date, datetime.time, datetime.datetime)


@dataclasses.dataclass
class MinimalInclusive(Constraint[T]):
    """
    A constraint that enforces a lower inclusive bound on a value.

    This constraint ensures that a given value is greater than or
    equal to the specified minimum inclusive bound. It can be applied
    to various types, including Integer, Real, Date, Time, and
    Timestamp. The constraint raises errors if the input value is
    less than the bound or if the input is of an incorrect type.

    Args:
      value: The lower inclusive limit for the value. The value must
        be of the same type as this attribute and must be greater
        than or equal to this limit.
    """

    value: T
    """
    The lower inclusive limit for the value. The value must be of
    the same type as this attribute and must be greater than or
    equal to this limit.
    """

    async def validate(self, value: T) -> bool:
        if not isinstance(value, type(self.value)):
            raise TypeError(f"Expected value of type '{type(self.value).__name__}', received '{type(value).__name__}'.")

        if not value >= self.value:
            raise ValueError(
                f"Value '{value}' must be greater than or equal to the minimal inclusive limit of '{self.value}'."
            )

        return True

    def serialize(self, serializer: Serializer) -> None:
        if isinstance(self.value, int):
            serializer.write_int("MinimalInclusive", self.value)
        elif isinstance(self.value, float):
            serializer.write_float("MinimalInclusive", self.value)
        elif isinstance(self.value, datetime.date):
            serializer.write_date("MinimalInclusive", self.value)
        elif isinstance(self.value, datetime.time):
            serializer.write_time("MinimalInclusive", self.value)
        elif isinstance(self.value, datetime.datetime):
            serializer.write_datetime("MinimalInclusive", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (Integer, Real, Date, Time, Timestamp)):
            raise ValueError(
                f"Expected constraint's data type to be 'Integer', 'Real', 'Date', 'Time' or 'Timestamp', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="MinimalInclusive")
        if issubclass(data_type, Integer):
            value = yield from deserializer.read_int()
        elif issubclass(data_type, Real):
            value = yield from deserializer.read_float()
        elif issubclass(data_type, Date):
            value = yield from deserializer.read_date()
        elif issubclass(data_type, Time):
            value = yield from deserializer.read_time()
        elif issubclass(data_type, Timestamp):
            value = yield from deserializer.read_datetime()
        yield from deserializer.read_end_element(name="MinimalInclusive")

        return cls(typing.cast(T, value))



import collections.abc
import dataclasses
import functools

import typing_extensions as typing

from ..fdl import Deserializer, Serializer
from ..identifiers import DataTypeIdentifier
from ..protobuf import DecodeError, Reader, WireType, Writer
from ..validators import check_display_name, check_identifier
from .any import Any
from .constrained import Constrained
from .convertable import Native
from .data_type import DataType
from .list import List

if typing.TYPE_CHECKING:
    from ..common import Context, Feature

T = typing.TypeVar("T", bound=Native)


@dataclasses.dataclass
class Custom(DataType[T]):
    """
    A user-defined data type that can be used through other data types.

    Attributes:
      identifier: Uniquely identifies the custom data type within the
        scope of the same feature.
      display_name: Human readable name of the custom data type.
      description: Describes the use and purpose of the custom data
        type.
      data_type: The SiLA data type of the custom data type.
      value: The custom SiLA data type instances.
    """

    identifier: typing.ClassVar[str] = ""
    display_name: typing.ClassVar[str] = ""
    description: typing.ClassVar[str] = ""
    data_type: typing.ClassVar[type[DataType]] = Any
    feature: typing.ClassVar[typing.Optional["Feature"]] = None

    value: DataType = dataclasses.field(default_factory=Any)

    @classmethod
    @functools.cache
    def fully_qualified_identifier(cls) -> DataTypeIdentifier:
        """Uniquely identifies the custom data type definition."""

        if cls.feature is None:
            msg = (
                f"Unable to access fully qualified identifier of unregistered Custom "
                f"'{cls.identifier}'. Add it to a feature first."
            )
            raise RuntimeError(msg)

        return DataTypeIdentifier.create(**cls.feature.fully_qualified_identifier._data, data_type=cls.identifier)

    @typing.override
    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[T] = None, /) -> typing.Self:
        return await cls(value=await cls.data_type.from_native(context, value)).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> T:
        await self.validate()

        return await self.value.to_native(context)

    @typing.override
    async def validate(self) -> typing.Self:
        if not isinstance(self.value, self.data_type):
            msg = f"Expected value of type '{self.data_type.__name__}', received '{type(self.value).__name__}'."
            raise TypeError(msg)

        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        flag = False
        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            pos = reader.cursor
            tag = reader.read_uint32()

            if tag >> 3 == 1:
                reader.expect_type(tag, WireType.LEN)

                try:
                    if issubclass(cls.data_type, List) or (
                        issubclass(cls.data_type, Constrained) and issubclass(cls.data_type.data_type, List)
                    ):
                        if not flag:
                            values = cls.data_type.decode(reader.buffer[pos:end])
                            message.value = values
                            flag = True

                        reader.skip_type(tag & 7)
                    else:
                        message.value = message.value.decode(reader, reader.read_uint32())
                except DecodeError as error:
                    msg = f"Invalid field '{cls.identifier}' in message '{cls.__name__}': {error.message}"
                    raise DecodeError(msg, error.offset, [cls.identifier, *error.path]) from None

            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        self.value.encode(writer, 1)

        if number:
            writer.ldelim()

        return writer.finish()

    @typing.override
    @classmethod
    def serialize(cls, serializer: Serializer, *, definition: bool = False) -> None:
        if definition:
            serializer.start_element("DataTypeDefinition")
            serializer.write_str("Identifier", cls.identifier)
            serializer.write_str("DisplayName", cls.display_name)
            serializer.write_str("Description", cls.description)
            cls.data_type.serialize(serializer)
            serializer.end_element("DataTypeDefinition")
        else:
            serializer.start_element("DataType")
            serializer.write_str("DataTypeIdentifier", cls.identifier)
            serializer.end_element("DataType")

    @typing.override
    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, *, definition: bool = False
    ) -> collections.abc.Generator[None, typing.Any, type[typing.Self]]:
        if definition:
            yield from deserializer.read_start_element(name="DataTypeDefinition")

            yield from deserializer.read_start_element(name="Identifier")
            identifier = yield from deserializer.read_str()
            check_identifier(identifier)
            yield from deserializer.read_end_element(name="Identifier")

            yield from deserializer.read_start_element(name="DisplayName")
            display_name = yield from deserializer.read_str()
            check_display_name(display_name)
            yield from deserializer.read_end_element(name="DisplayName")

            yield from deserializer.read_start_element(name="Description")
            description = yield from deserializer.read_str()
            yield from deserializer.read_end_element(name="Description")

            data_type = yield from deserializer.read(DataType.deserialize(deserializer))
            yield from deserializer.read_end_element(name="DataTypeDefinition")

            return cls.create(identifier, display_name, description, data_type)
        else:
            yield from deserializer.read_start_element(name="DataTypeIdentifier")

            identifier = yield from deserializer.read_str()
            check_identifier(identifier)

            yield from deserializer.read_end_element(name="DataTypeIdentifier")

            return cls.create(identifier=identifier, display_name=identifier)

    @classmethod
    def add_to_feature(cls, feature: "Feature") -> type[typing.Self]:
        """
        Register this custom data type with a feature.

        Args:
          feature: The feature to add this custom data type to.

        Returns:
          The class, allowing for method chaining.
        """

        feature.data_type_definitions[cls.identifier] = cls

        return cls

    @classmethod
    def create(
        cls,
        identifier: str,
        display_name: str,
        description: str = "",
        data_type: type[DataType] = Any,
        feature: typing.Optional["Feature"] = None,
    ) -> type[typing.Self]:
        """
        Create a new SiLA `Custom` class with the provided data type.

        Args:
          identifier: Uniquely identifies the custom data type within the
            scope of the same feature.
          display_name: Human readable name of the custom data type.
          description: Describes the use and purpose of the custom data
            type.
          data_type: The SiLA data type for the custom value.
          feature: The feature custom data type is assigned to.
          name: An optional name for the new `Custom` class.

        Returns:
          A new `Custom` class with the specified data type.
        """

        check_identifier(identifier)
        check_display_name(display_name)

        custom: type[typing.Self] = dataclasses.make_dataclass(
            identifier or cls.__name__,
            [("value", data_type, dataclasses.field(default_factory=data_type))],
            bases=(cls,),
            namespace={
                "__doc__": description,
                "identifier": identifier,
                "display_name": display_name,
                "description": description,
                "data_type": data_type,
                "feature": feature,
            },
            eq=False,
        )

        if feature is not None:
            custom.add_to_feature(feature)

        return custom

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Custom):
            return NotImplemented

        return (
            self.identifier == other.identifier
            and self.display_name == other.display_name
            and self.description == other.description
            and self.data_type.__name__ == other.data_type.__name__
            and self.value == other.value
        )



import grpc.aio
import typing_extensions as typing

from .channel_options import ChannelOptions


async def create_server(
    address: str,
    /,
    tls: bool = True,
    require_client_auth: bool = False,
    root_certificates: typing.Optional[bytes] = None,
    certificate_chain: typing.Optional[bytes] = None,
    private_key: typing.Optional[bytes] = None,
    options: typing.Optional[ChannelOptions] = None,
    **kwargs,
) -> tuple[grpc.aio.Server, int]:
    """
    Create a server at the address.

    Args:
      address: The target address to bind to.
      require_client_auth: A boolean indicating whether or not to
        require clients to be authenticated. May only be True if
        root_certificates is not None.
      tls: Whether to run a secure/TLS channel or a plaintext channel
        (i.e. no TLS), defaults to run with TLS encryption.
      root_certificates: The PEM-encoded root certificates as a byte
        string, or None to retrieve them from a default location
        chosen by gRPC runtime.
      certificate_chain: The PEM-encoded certificate chain as a byte
        string to use or None if no certificate chain should be used.
      private_key: The PEM-encoded private key as a byte string, or
        None if no private key should be used.
      options: Additional options for the underlying gRPC connection.

    Returns:
      The grpc channel connected to cloud client endpoint.
    """

    options = ChannelOptions(options or {})

    server = grpc.aio.server()

    if tls:
        credentials = grpc.ssl_server_credentials(
            private_key_certificate_chain_pairs=[(private_key, certificate_chain)]
            if private_key is not None and certificate_chain is not None
            else [],
            root_certificates=root_certificates,
            require_client_auth=require_client_auth,
        )
        port = server.add_secure_port(address, credentials)
    else:
        port = server.add_insecure_port(address)

    return server, port



import base64
import collections.abc
import contextlib
import re
import unittest.mock

import grpc.aio
import typing_extensions as typing

from sila.framework.binary_transfer import BinaryTransferError
from sila.framework.errors import SiLAError


@contextlib.contextmanager
def raises(
    expected_exception: type["SiLAError"], match: typing.Union[None, str, re.Pattern] = None
) -> collections.abc.Iterator[grpc.aio.ServicerContext]:
    """
    Assert that a grpc handler raises an exception type.

    Examples:
      Use `raises` as a context manager, which will capture the
      exception of the given type, or any of its subclasses:
      >>> from sila import SiLAError
      >>> from sila.testing import raises
      >>> with raises(SiLAError) as context:
      >>>   await my_grpc_handler(request_message, context)

      If the grpc handler does not raise the expected exception, or
      no exception at all, the check will fail instead.

      You can also use the keyword argument match to assert that the
      exception matches a text or regex:
      >>> from sila import SiLAError
      >>> from sila.testing import raises
      >>> with raises(SiLAError, match=r"must be 0 or None") as context:
      >>>   await my_grpc_handler(request_message, context)

    Args:
      expected_exception: The expected exception type. Note that
        subclasses of the passed exceptions will also match.
      match: If specified, a string containing a regular expression,
        or a regular expression object, that is tested against the
        string representation of the exception

    Yields:
      A mocked grpc servicer context passed into the grpc handler.
    """

    abort = unittest.mock.AsyncMock()

    with contextlib.suppress(expected_exception, StopAsyncIteration):
        yield unittest.mock.Mock(
            spec=grpc.aio.ServicerContext, abort=abort, invocation_metadata=unittest.mock.Mock(return_value=())
        )

    if abort.call_count != 1:
        msg = f"Did not raise {expected_exception}"
        raise AssertionError(msg)

    code: grpc.StatusCode = abort.call_args.kwargs["code"]
    if code != grpc.StatusCode.ABORTED:
        msg = f"Expected status code to be 'aborted', received '{code.value[1]}'."
        raise AssertionError(msg)

    if match is None:
        return

    details: str = abort.call_args.kwargs["details"]
    buffer = base64.b64decode(details)

    if issubclass(expected_exception, BinaryTransferError):
        message = BinaryTransferError.decode(buffer)
    elif issubclass(expected_exception, SiLAError):
        message = SiLAError.decode(buffer)

    pattern = re.compile(match)
    if not pattern.match(message.message):
        msg = f"Regex pattern did not match.\n Regex: {match!r}\n Input: {message.message!r}"
        raise AssertionError(msg)



import collections.abc
import dataclasses

import typing_extensions as typing

from ..fdl import Deserializer, Serializer
from .data_type import BasicType, DataType
from .string import String

if typing.TYPE_CHECKING:
    from ..common import Context
    from ..fdl import Serializer
    from ..protobuf import Reader, Writer


@dataclasses.dataclass
class Void(BasicType):
    """Represents no data as a value of the SiLA any type."""

    @typing.override
    @classmethod
    async def from_native(cls, context: "Context", value: None = None, /) -> typing.Self:
        return await cls().validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> None:
        await self.validate()

        return

    @typing.override
    async def validate(self) -> typing.Self:
        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union["Reader", bytes], length: typing.Optional[int] = None) -> typing.Self:
        return cls()

    @typing.override
    def encode(self, writer: typing.Optional["Writer"] = None, number: typing.Optional[int] = None) -> bytes:
        return b""

    @typing.override
    @classmethod
    def serialize(cls, serializer: "Serializer") -> None:
        serializer.start_element("DataType")
        serializer.start_element("Constrained")
        serializer.start_element("DataType")
        serializer.write_str("Basic", "String")
        serializer.end_element("DataType")
        serializer.start_element("Constraints")
        serializer.write_int("Length", 0)
        serializer.end_element("Constraints")
        serializer.end_element("Constrained")
        serializer.end_element("DataType")

    @typing.override
    @classmethod
    def deserialize(cls, deserializer: Deserializer) -> collections.abc.Generator[None, typing.Any, type["Void"]]:
        from ..constraints import Length

        yield from deserializer.read_start_element(name="Constrained")
        data_type = yield from deserializer.read(DataType.deserialize(deserializer))
        assert data_type is String

        yield from deserializer.read_start_element(name="Constraints")
        yield from deserializer.read(Length.deserialize(deserializer, data_type))
        yield from deserializer.read_end_element("Constraints")
        yield from deserializer.read_end_element(name="Constrained")

        return cls

    @typing.override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Void):
            return NotImplemented

        return True



import weakref

import typing_extensions as typing

from ..errors import InvalidCommandExecutionUUID
from ..protobuf import Protobuf

if typing.TYPE_CHECKING:
    from ..binary_transfer import BinaryTransferHandler
    from ..command import CommandExecution
    from ..identifiers import FeatureIdentifier
    from ..protobuf import Protobuf
    from .feature import Feature


class Context:
    """Mediates access to various functionalities within the library."""

    def __init__(self) -> None:
        self._protobuf: Protobuf = Protobuf(self)
        self._features: dict["FeatureIdentifier", "Feature"] = {}
        self._command_executions: dict[str, "CommandExecution"] = {}
        self._binary_transfer_handler: "BinaryTransferHandler"

    @property
    def protobuf(self) -> "Protobuf":
        """A collection of protobuf messages and services."""

        return self._protobuf

    @property
    def features(self) -> dict["FeatureIdentifier", "Feature"]:
        """A collection of registered features."""

        return self._features

    @property
    def command_executions(self) -> dict[str, "CommandExecution"]:
        """A collection of currently executed commands."""

        return self._command_executions

    @property
    def binary_transfer_handler(self) -> "BinaryTransferHandler":
        """Upload and download large binaries in chunks."""

        return self._binary_transfer_handler

    def register_feature(self, feature: "Feature") -> None:
        """
        Register a feature to the context.

        Args:
          feature: The feature to register.
        """

        self.protobuf.merge(feature.context.protobuf)
        self.features[feature.fully_qualified_identifier] = feature
        feature.context = weakref.proxy(self)

    def add_command_execution(self, command_execution: "CommandExecution") -> None:
        """
        Add a command execution to the context.

        Args:
          command_execution: The command execution to add.

        Raises:
          InvalidCommandExecutionUUID: If the uuid is already registered
            with a command execution.
        """

        if command_execution.command_execution_uuid in self.command_executions:
            msg = f"Command execution with uuid '{command_execution.command_execution_uuid}' already exists."
            raise InvalidCommandExecutionUUID(msg)

        self.command_executions[command_execution.command_execution_uuid] = command_execution

    def get_command_execution(self, command_execution_uuid: str) -> "CommandExecution":
        """
        Get a command execution by its UUID.

        Args:
          command_execution_uuid: The UUID of the command execution.

        Returns:
          The command execution with the given uuid.

        Raises:
          InvalidCommandExecutionUUID: If no command execution is
            available for the uuid.
        """

        if command_execution_uuid not in self.command_executions:
            msg = f"Requested unknown command execution uuid '{command_execution_uuid}'."
            raise InvalidCommandExecutionUUID(msg)

        return self.command_executions[command_execution_uuid]



import io
import struct

from .encode_error import EncodeError


class Writer:
    """
    A data sink to which an encoded Protocol Buffer message can be
    written.
    """

    def __init__(self):
        self.buf = [io.BytesIO()]
        self.len = 0

    @property
    def buffer(self) -> io.BytesIO:
        """
        The internal buffer that holds the provided byte data.
        """

        return self.buf[-1]

    @property
    def length(self) -> int:
        """
        The total length of the byte buffer.
        """

        return self.len

    def write_varint(self, value: int) -> "Writer":
        """
        Encode a variable-length integer (varint) in Protocol Buffer
        format and writes it to the buffer.

        Args:
          value: The integer value to encode as a varint. Must be
            non-negative.

        Returns:
          The Writer instance, allowing for method chaining.

        Raises:
          EncodeError: If the input value is negative, as varint encoding
            does not support negative integers.
        """

        if value < 0:
            raise EncodeError("Varint encoding does not support negative values.")

        while value > 0x7F:
            # Write 7 bits and set MSB to 1.
            self.buffer.write(bytes((value & 0x7F | 0x80,)))
            # Shift value right by 7 bits.
            value >>= 7
            self.len += 1

        # Write the final 7 bits (MSB = 0).
        self.buffer.write(bytes((value,)))
        self.len += 1

        return self

    def write_uint32(self, value: int) -> "Writer":
        """
        Encode an unsigned 32-bit integer as a varint and writes it to
        the buffer.

        Args:
          value: The unsigned 32-bit integer to encode and write. Must be
          in the range [0, 2^32 - 1].

        Returns:
          The Writer instance, allowing for method chaining.

        Raises:
          EncodeError: If the value is not within the range of an unsigned
            32-bit integer.
        """

        if not (0 <= value < (1 << 32)):
            raise EncodeError("Value must be a 32-bit unsigned integer (0 <= value < 2^32).")

        return self.write_varint(value)

    def write_int32(self, value: int) -> "Writer":
        """
        Encode a signed 32-bit integer as a varint and writes it to the
        buffer.

        Args:
          value: The signed 32-bit integer to encode and write. Must be
            in the range [-2^31, 2^31 - 1].

        Returns:
          The Writer instance, allowing for method chaining.

        Raises:
          EncodeError: If the value is not within the range of a signed
            32-bit integer.
        """

        if not (-(1 << 31) <= value < (1 << 31)):
            raise EncodeError("Value must be a 32-bit signed integer (-2^31 <= value < 2^31).")

        # Map negative values to the unsigned space using two's complement
        if value < 0:
            value += 1 << 64

        return self.write_varint(value)

    def write_uint64(self, value: int) -> "Writer":
        """
        Encode an unsigned 64-bit integer as a varint and writes it to
        the buffer.

        Args:
          value: The unsigned 64-bit integer to encode and write. Must be
            in the range [0, 2^64 - 1].

        Returns:
          The Writer instance, allowing for method chaining.

        Raises:
          EncodeError: If the value is not within the range of an unsigned
            64-bit integer.
        """

        if not (0 <= value < (1 << 64)):
            raise EncodeError("Value must be a 64-bit unsigned integer (0 <= value < 2^64).")

        return self.write_varint(value)

    def write_int64(self, value: int) -> "Writer":
        """
        Encode a signed 64-bit integer as a varint and writes it to the
        buffer.

        Args:
          value: The signed 64-bit integer to encode and write. Must be
            in the range [-2^63, 2^63 - 1].

        Returns:
          The Writer instance, allowing for method chaining.

        Raises:
          EncodeError: If the value is not within the range of a signed
            64-bit integer.
        """

        if not (-(1 << 63) <= value < (1 << 63)):
            raise EncodeError("Value must be a 64-bit signed integer (-2^63 <= value < 2^63).")

        # Map negative values to the unsigned space using two's complement
        if value < 0:
            value += 1 << 64

        return self.write_varint(value)

    def write_double(self, value: float) -> "Writer":
        """
        Encode a 64-bit double-precision floating-point number (IEEE
        754) and writes it to the buffer.

        Args:
          value: The double-precision floating-point number to encode and
            write.

        Returns:
          The Writer instance, allowing for method chaining.

        Raises:
          ValueError: If the value is not within the range of a 64-bit
            floating-point number.
        """

        try:
            # Encode the float as a fixed 8-byte little-endian representation
            buffer = struct.Struct("<d").pack(value)
        except struct.error:
            raise EncodeError("Value must be a 64-bit double-precision floating-point number.")

        self.buffer.write(buffer)
        self.len += 8

        return self

    def write_bool(self, value: bool) -> "Writer":
        """
        Encode a boolean value as a single byte and writes it to the
        buffer.

        Args:
          value: The boolean value to encode and write.

        Returns:
          The Writer instance, allowing for method chaining.
        """

        self.buffer.write(b"\x01" if value else b"\x00")
        self.len += 1

        return self

    def write_bytes(self, value: bytes) -> "Writer":
        """
        Encode a byte sequence with a length prefix and writes it to the
        buffer.

        Args:
          value: The byte sequence to encode and write.

        Returns:
          The Writer instance, allowing for method chaining.
        """

        length = len(value)

        # Write the length prefix as a varint
        self.write_varint(length)

        # Write the byte sequence (if non-empty)
        if length > 0:
            self.buffer.write(value)
            self.len += length

        return self

    def write_string(self, value: str) -> "Writer":
        """
        Encode a UTF-8 string with a length prefix and writes it to the
        buffer.

        Args:
          value: The string to encode and write.

        Returns:
          The Writer instance, allowing for method chaining.
        """

        return self.write_bytes(value.encode("utf-8"))

    def write(self, value: bytes) -> "Writer":
        """
        Write raw bytes directly to the buffer.

        Args:
          value: The byte sequence to write.

        Returns:
          The Writer instance, allowing for method chaining.
        """

        self.buffer.write(value)
        self.len += len(value)

        return self

    def fork(self) -> "Writer":
        """
        Create a new fork within the buffer.

        Returns:
          The Writer instance, allowing for method chaining.
        """

        self.buf.append(io.BytesIO())

        return self

    def ldelim(self) -> "Writer":
        """
        Prefix the current fork with its length and writes it into the
        buffer.

        Returns:
          The Writer instance, allowing for method chaining.
        """

        fork = self.buf.pop()
        value = fork.getvalue()
        length = len(value)

        self.write_varint(length)
        self.buffer.write(value)

        fork.close()

        return self

    def finish(self) -> bytes:
        """
        Finalize the writer and retrieves the serialized data.

        Returns:
          The serialized data from the writer's buffer.
        """

        value = self.buffer.getvalue()

        return value



import collections.abc
import dataclasses

import typing_extensions as typing

from ..data_types.binary import Binary
from ..data_types.string import String
from ..fdl import Deserializer, Serializer
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", str, bytes)


@dataclasses.dataclass
class MaximalLength(Constraint[T]):
    """
    A constraint that enforces a maximum length for a value of type
    String or Binary.

    This constraint is used to ensure that a value does not exceed a
    specified maximum length. It raises errors if the value is not of
    the correct type or if the length exceeds the defined limit. The
    maximum length must be a positive integer and less than 2⁶³.

    Args:
      maximal_length: The maximum allowed length for the value. It
        must be a non-negative integer and less than 2⁶³.

    Raises:
      ValueError: If the maximum length (`value`) is negative or
        exceeds 2⁶³.
    """

    value: int
    """
    The maximum allowed length for the value. It must be a
    non-negative integer and less than 2⁶³.
    """

    def __post_init__(self):
        if self.value < 0:
            raise ValueError(f"Maximal length must be a non-negative integer, received '{self.value}'.")
        if self.value >= 2**63:
            raise ValueError(f"Maximal length must be less than 2⁶³, received '{self.value}'.")

    async def validate(self, value: T) -> bool:
        if not isinstance(value, (str, bytes)):
            raise TypeError(f"Expected value of type 'str' or 'bytes', received '{type(value).__name__}'.")

        if len(value) > self.value:
            raise ValueError(f"Expected value with maximal length '{self.value}', received '{value}'.")

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.write_int("MaximalLength", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (String, Binary)):
            raise ValueError(
                f"Expected constraint's data type to be 'String' or 'Binary', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="MaximalLength")
        value = yield from deserializer.read_int()
        yield from deserializer.read_end_element(name="MaximalLength")

        return cls(value)
